#!/usr/bin/env python3

import boto3
import logging
import sys
import json
import os, re
from botocore.exceptions import ClientError
from datetime import datetime
from pyawscp.Functions import Functions
from pyawscp.Utils import Utils, Style
from pyawscp.PrettyTable import PrettyTable
from pyawscp.Emoticons import Emoticons
from pyawscp.Config import Config
from pyawscp.CsvDrawIo import CsvDrawIo
from pyawscp.TableArgs import TableArgs
from pyawscp.CsvDrawIo import CsvDrawIo
from pygments import highlight, lexers, formatters

LOG = logging.getLogger("app." + __name__)
LOG.setLevel(logging.INFO)

class PyEc2NetSecCP:

    def __init__(self, config):
        self.config = config
        self.tempCredentials = None

    
    # List all Security Groups
    def listSecurityGroup(self):
        ec2 = self.botoSession().client('ec2')

        tableArgs = TableArgs()
        tableArgs.setArguments(self.config.commandArguments)

        listResourcesAssociated = False
        listIpPermissions       = False
        
        if "list-associated" in self.config.commandArguments:
           listResourcesAssociated    = True
        if "list-permissions" in self.config.commandArguments:
           listIpPermissions          = True 

        filters = []
        if self.config.awsTagsToFilter():
           # Tags from Environment
           for tag in self.config.awsTags: 
               filters.append(
                   {'Name':'tag:' + tag,'Values':[self.config.awsTags[tag]]}
               )

        # Tags from command line arguments
        if len(tableArgs.tagsTemp) > 0:
           for tag in tableArgs.tagsTemp: 
               filters.append(
                   {'Name':'tag:' + tag,'Values':[tableArgs.tagsTemp[tag]]}
               )
        securityGroups = ec2.describe_security_groups(Filters=filters)

        jsonResult = ""
        if self.config.printResults or tableArgs.verbose or tableArgs.saveToFile:
           jsonResult = Utils.dictToJson(securityGroups)

        resultTxt = ""
        if len(securityGroups["SecurityGroups"]) >= 1:
            idx_lin             = 0
            data                = {}
            data["sgs"]         = []
            ipPermissions       = []
            ipPermissionsEgress = []
            for securityGroup in securityGroups["SecurityGroups"]:
                idx_lin += 1
                securityGroupId          = securityGroup['GroupId']
                tagsSg                   = securityGroup['Tags'] if "Tags" in securityGroup else None
                networkingInterfacesData = []

                # Collect Network Interfaces associated with the Security Group
                if listResourcesAssociated:

                   filterNetworkInterfaces = [{
                      'Name':'group-id','Values':[securityGroupId]
                   }]
                   networkInterfaces = ec2.describe_network_interfaces(Filters=filterNetworkInterfaces)
                   if len(networkInterfaces) > 0:
                      for ni in networkInterfaces["NetworkInterfaces"]:
                          networkingInterfacesData.append({
                              'NetworkInterfaceId': ni['NetworkInterfaceId'],
                              'Description': ni['Description'],
                              'PrivateIpAddress': ni['PrivateIpAddress'],
                              'VpcId': ni['VpcId']
                           })
               
                if listIpPermissions:
                   if len(securityGroup["IpPermissions"]) > 0:
                      ipPermissions = []
                      for ipp in securityGroup["IpPermissions"]:
                          ipRanges = []
                          for ipRange in ipp["IpRanges"]:
                             ipRanges.append({
                                "cidrIp": ipRange["CidrIp"],
                                "description": ipRange["Description"] if "Description" in ipRange else None
                             }) 
                          ipPermissions.append({
                            "fromPort": ipp["FromPort"] if "FromPort" in ipp else None,
                            "toPort": ipp["ToPort"] if "ToPort" in ipp else None,
                            "ipProtocol": ipp["IpProtocol"],
                            "ipRanges": ipRanges
                          })

                   if len(securityGroup["IpPermissionsEgress"]) > 0:
                      ipPermissionsEgress = []
                      for ippEgress in securityGroup["IpPermissionsEgress"]:
                          ipRangesEgress = []
                          for ipRangeEgress in ippEgress["IpRanges"]:
                             ipRangesEgress.append({
                                "cidrIp": ipRangeEgress["CidrIp"],
                                "description": ipRangeEgress["Description"] if "Description" in ipRangeEgress else None
                             }) 
                          ipPermissionsEgress.append({
                            "ipProtocol": ippEgress["IpProtocol"],
                            "ipRanges": ipRangesEgress
                          })

                data["sgs"].append({
                   "sg": {
                      "index": idx_lin,
                      "id":securityGroupId,
                      "vpcId": securityGroup['VpcId'],
                      "networkInterfaces": networkingInterfacesData,
                      "ipPermissions": ipPermissions if ipPermissions else [],
                      "ipPermissionsEgress": ipPermissionsEgress if ipPermissionsEgress else [],
                      "tags": tagsSg
                   }
                })
            
            #print(Utils.saveDictAsJson("TESTSSGs",data))

            ## Print Report from Data Collected
            resultTxt = "List of Security Groups, total of " + Style.GREEN + str(len(securityGroups["SecurityGroups"])) + Style.RESET + "\n"
            resultTxt = resultTxt + Utils.separator()  + "\n\n"
            
            header = ["#","SG Id", "VPC"]
            if listResourcesAssociated:
               header.append("Network Intf")
               header.append("Description")
               header.append("Private IP")
            if listIpPermissions:
               header.append("I-Prot | I-PortRange | I-Source             | I-Description                    (Ingress)")
               header.append("E-Prot | E-Source             | E-Description                    (Egress)")
            if tableArgs.showTags:
               header.append("SG Tags")       

            prettyTable = PrettyTable(header)
            for securityGroup in data["sgs"]:

                securityGroupId = securityGroup["sg"]["id"]
                vpc             = securityGroup["sg"]["vpcId"] 
                columns         = [ securityGroup["sg"]["index"], securityGroupId, vpc ]

                if listResourcesAssociated and not listIpPermissions:
                   for ni in securityGroup["sg"]["networkInterfaces"]:
                       columns.append(ni["NetworkInterfaceId"])
                       columns.append(ni["Description"] if ni["Description"] != "" else "EC2" )
                       columns.append(ni["PrivateIpAddress"])
                       
                       if tableArgs.showTags:
                          tags = Utils.formatPrintTags( securityGroup["sg"]["tags"])
                          columns.append(tags)

                       prettyTable.addRow(columns)
                       columns = [ securityGroup["sg"]["index"], securityGroupId, vpc ]
                   prettyTable.addSeparatorGroup()

                elif not listResourcesAssociated and listIpPermissions:
                   for index, val in enumerate(securityGroup["sg"]["ipPermissions"]):
                       protocol = "All" if val["ipProtocol"] == -1 else str(val["ipProtocol"]).upper()
                       if len(val["ipRanges"]) > 0:
                           for ip in val["ipRanges"]:

                                 desc = ip["description"] if ip["description"] else "" 
                                 desc = (desc[:38] + '..') if len(desc) > 38 else desc

                                 if val["fromPort"] != val["toPort"]:
                                    ports = (str(val["fromPort"]) + "-" + str(val["toPort"])).ljust(11," ")
                                 else:
                                    ports = str(val["fromPort"]).ljust(11," ")

                                 line = str(protocol).upper().ljust(6," ") + Style.RESET + " | " + Style.GREEN + \
                                        ports + Style.RESET + " | " + Style.GREEN + \
                                        str(ip["cidrIp"] ).ljust(20," ") + Style.RESET + " | " + Style.GREEN + \
                                        desc
                                 columns.append(line)
                                 columns.append("")
                                 if tableArgs.showTags:
                                    tags = Utils.formatPrintTags( securityGroup["sg"]["tags"])
                                    columns.append(tags)
                                 prettyTable.addRow(columns)
                                 columns = [ securityGroup["sg"]["index"], securityGroupId, vpc ]
                       else:
                           line = str(protocol).upper().ljust(6," ") + Style.RESET + " | " + Style.GREEN + \
                                  str(val["fromPort"] ).ljust(5," ")
                           columns.append(line)
                           columns.append("")
                           if tableArgs.showTags:
                              tags = Utils.formatPrintTags( securityGroup["sg"]["tags"])
                              columns.append(tags)
                           prettyTable.addRow(columns)
                           columns = [ securityGroup["sg"]["index"], securityGroupId, vpc ]

                   for index, val in enumerate(securityGroup["sg"]["ipPermissionsEgress"]):
                       protocol = "All" if val["ipProtocol"] == -1 else str(val["ipProtocol"]).upper()
                       if len(val["ipRanges"]) > 0:
                           for ip in val["ipRanges"]:
                                 desc = ip["description"] if ip["description"] else "" 
                                 desc = (desc[:38] + '..') if len(desc) > 38 else desc

                                 line = str(protocol).ljust(6," ") + Style.RESET + " | " + Style.GREEN + \
                                       str(ip["cidrIp"] ).ljust(20," ") + Style.RESET + " | " + Style.GREEN + \
                                       desc

                                 columns.append(" ")   
                                 columns.append(line)

                                 if tableArgs.showTags:
                                    tags = Utils.formatPrintTags( securityGroup["sg"]["tags"])
                                    columns.append(tags)
                                 prettyTable.addRow(columns)
                                 columns = [ securityGroup["sg"]["index"], securityGroupId, vpc ]
                       else:
                           columns.append("")
                           line = str(protocol).ljust(6," ")
                           columns.append(line)
                           
                           if tableArgs.showTags:
                              tags = Utils.formatPrintTags( securityGroup["sg"]["tags"])
                              columns.append(tags)
                           prettyTable.addRow(columns)
                           columns = [ securityGroup["sg"]["index"], securityGroupId, vpc ]

                   prettyTable.addSeparatorGroup() 

                elif listResourcesAssociated and listIpPermissions:
                   for ni in securityGroup["sg"]["networkInterfaces"]:
                       columns = [ securityGroup["sg"]["index"], securityGroupId, vpc ]
                       columns.append(ni["NetworkInterfaceId"])
                       columns.append(ni["Description"] if ni["Description"] != "" else "EC2" )
                       columns.append(ni["PrivateIpAddress"])

                       for index, val in enumerate(securityGroup["sg"]["ipPermissions"]):
                           protocol = "All" if val["ipProtocol"] == -1 else str(val["ipProtocol"]).upper()
                           if len(val["ipRanges"]) > 0:
                                 for ip in val["ipRanges"]:

                                       desc = ip["description"] if ip["description"] else "" 
                                       desc = (desc[:38] + '..') if len(desc) > 38 else desc

                                       if val["fromPort"] != val["toPort"]:
                                          ports = (str(val["fromPort"]) + "-" + str(val["toPort"])).ljust(11," ")
                                       else:
                                          ports = str(val["fromPort"]).ljust(11," ")

                                       line = str(protocol).ljust(6," ") + Style.RESET + " | " + Style.GREEN + \
                                             ports + Style.RESET + " | " + Style.GREEN +\
                                             str(ip["cidrIp"] ).ljust(20," ") + Style.RESET + " | " + Style.GREEN + \
                                             desc
                                       columns.append(line)
                                       columns.append("")
                                       if tableArgs.showTags:
                                          tags = Utils.formatPrintTags( securityGroup["sg"]["tags"])
                                          columns.append(tags)
                                       prettyTable.addRow(columns)
                                       columns = [ securityGroup["sg"]["index"], securityGroupId, vpc ]
                                       columns.append(ni["NetworkInterfaceId"])
                                       columns.append(ni["Description"] if ni["Description"] != "" else "EC2" )
                                       columns.append(ni["PrivateIpAddress"])
                           else:
                                 line = str(protocol).ljust(6," ") + Style.RESET + " | " + Style.GREEN + \
                                       str(val["fromPort"] ).ljust(5," ")
                                 columns.append(line)
                                 columns.append("")
                                 if tableArgs.showTags:
                                    tags = Utils.formatPrintTags( securityGroup["sg"]["tags"])
                                    columns.append(tags)
                                 prettyTable.addRow(columns)
                                 columns = [ securityGroup["sg"]["index"], securityGroupId, vpc ]
                                 columns.append(ni["NetworkInterfaceId"])
                                 columns.append(ni["Description"] if ni["Description"] != "" else "EC2" )
                                 columns.append(ni["PrivateIpAddress"])
                        
                       for index, val in enumerate(securityGroup["sg"]["ipPermissionsEgress"]):
                           protocol = "All" if val["ipProtocol"] == -1 else str(val["ipProtocol"]).upper()
                           if len(val["ipRanges"]) > 0:
                                 for ip in val["ipRanges"]:
                                       desc = ip["description"] if ip["description"] else "" 
                                       desc = (desc[:38] + '..') if len(desc) > 38 else desc

                                       line = str(protocol).ljust(6," ") + Style.RESET + " | " + Style.GREEN + \
                                             str(ip["cidrIp"] ).ljust(20," ") + Style.RESET + " | " + Style.GREEN + \
                                             desc

                                       columns.append(" ")   
                                       columns.append(line)

                                       if tableArgs.showTags:
                                          tags = Utils.formatPrintTags( securityGroup["sg"]["tags"])
                                          columns.append(tags)
                                       prettyTable.addRow(columns)
                                       columns = [ securityGroup["sg"]["index"], securityGroupId, vpc ]
                                       columns.append(ni["NetworkInterfaceId"])
                                       columns.append(ni["Description"] if ni["Description"] != "" else "EC2" )
                                       columns.append(ni["PrivateIpAddress"])
                           else:
                                 columns.append("")
                                 line = str(protocol).ljust(6," ")
                                 columns.append(line)
                                 
                                 if tableArgs.showTags:
                                    tags = Utils.formatPrintTags( securityGroup["sg"]["tags"])
                                    columns.append(tags)
                                 prettyTable.addRow(columns)
                                 columns = [ securityGroup["sg"]["index"], securityGroupId, vpc ]
                                 columns.append(ni["NetworkInterfaceId"])
                                 columns.append(ni["Description"] if ni["Description"] != "" else "EC2" )
                                 columns.append(ni["PrivateIpAddress"])
                       
                       # Separator
                       columns = [ securityGroup["sg"]["index"], securityGroupId, vpc ]
                       columns.append("-".ljust(20,"-"))
                       columns.append("-".ljust(50,"-"))
                       columns.append("-".ljust(14,"-"))
                       columns.append("-".ljust(70,"-"))
                       columns.append("-".ljust(70,"-"))
                       if tableArgs.showTags:
                          columns.append("-".ljust(30,"-"))
                       prettyTable.addRow(columns)
                   
                   prettyTable.addSeparatorGroup()

                else:
                   if tableArgs.showTags:
                      tags = Utils.formatPrintTags( securityGroup["sg"]["tags"] )
                      columns.append(tags)
                   prettyTable.addRow(columns)

            if (int(tableArgs.sortCol) - 1) > len(columns):
                sortCol = "1"

            prettyTable.sortByColumn(int(tableArgs.sortCol) - 1)
            prettyTable.ascendingOrder(not tableArgs.desc)

            if tableArgs.csvDrawIo:
               csvDrawIo = CsvDrawIo(self.config)
               file = csvDrawIo.generateListSg('listSg_template.csv', prettyTable.header, data)
               print("")
               print("  --> " + Style.IGREEN + file + Style.GREEN + " generated!" + Style.RESET)
               print("")

            result = resultTxt + prettyTable.printMe("listSgs",self.config.tableLineSeparator, tableArgs)
            return Utils.formatResult(Functions.FUNCTIONS[Functions.LIST_SG]["name"], result, self.config, jsonResult, True, tableArgs)    
        else:
            return Utils.formatResult(Functions.FUNCTIONS[Functions.LIST_SG]["name"], " No Security Group found " + Emoticons.ops(), self.config, jsonResult, True, tableArgs)


    def botoSession(self):
        return self.config.botoSession()




import json
import os
import sys
from io import StringIO
import clipboard
from datetime import datetime
from pyawscp.Emoticons import Emoticons
from pygments import highlight, lexers, formatters

class Style:
    UNDERLINE = '\033[4m'
    RESET     = '\033[0m'

    # Regular Colors
    BLACK     = '\033[30m'
    RED       = '\033[31m'
    GREEN     = '\033[32m'
    YELLOW    = '\033[33m'
    BLUE      = '\033[34m'
    MAGENTA   = '\033[35m'
    CYAN      = '\033[36m'
    WHITE     = '\033[37m'

    # Bold
    BBLACK    ="\033[30m"
    BRED      ="\033[31m"
    BGREEN    ="\033[32m"
    BYELLOW   ="\033[33m"
    BBLUE     ="\033[34m"
    BMAGENTA  ="\033[35m"
    BCYAN     ="\033[36m"
    BWHITE    ="\033[37m"

    # Background Colors
    BG_BLACK  = '\033[40m'
    BG_RED    = '\033[41m'
    BG_GREEN  = '\033[42m'
    BG_YELLOW = '\033[43m'
    BG_BLUE   = '\033[44m'
    BG_PURPLE = '\033[45m'
    BG_CYAN   = '\033[46m'
    BG_WHITE  = '\033[47m'

    # High Intensty
    IBLACK    ="\033[90m"
    IRED      ="\033[91m"
    IGREEN    ="\033[92m"
    IYELLOW   ="\033[93m"
    IBLUE     ="\033[94m"
    IMAGENTA  ="\033[95m"
    ICYAN     ="\033[96m"
    IWWHITE   ="\033[97m"

    # Bold High Intensty
    BIBLACK   ="\033[90m"
    BIRED     ="\033[91m"
    BIGREEN   ="\033[92m"
    BIYELLOW  ="\033[93m"
    BIBLUE    ="\033[94m"
    BIPURPLE  ="\033[95m"
    BICYAN    ="\033[96m"
    BIWHITE   ="\033[97m"

    # High Intensty backgrounds
    On_IBLACK ="\033[100m"
    On_IRED   ="\033[101m"
    On_IGREEN ="\033[102m"
    On_IYELLOW="\033[103m"
    On_IBLUE  ="\033[104m"
    On_IPURPLE="\033[;95m"
    On_ICYAN  ="\033[106m"
    On_IWHITE ="\033[107m"

REMAIN_START = 72

class Utils:

    @staticmethod 
    def saveToFile(fileName, response) :
        if not os.path.exists("results"):
            os.makedirs("results")
        with open("results/" + fileName, 'w') as fout:
            json.dump(response, fout, indent=4)

    @staticmethod 
    def saveJsonToFile(fileName, jsonContent) :
        if not os.path.exists("results"):
            os.makedirs("results")
        with open("results/" + fileName, 'w') as fout:
            fout.write(jsonContent)
            fout.flush()
            print("OK Save")

    @staticmethod
    def saveDictAsJson(name, dict):
        if os.path.exists(name):
           os.delete(name)
        f = open(name + ".json", "a")
        f.write(json.dumps(dict))
        f.close()

    @staticmethod
    def loadJsonFile(fileName) :
        return json.load(open(fileName, 'r'))

    @staticmethod 
    def separator():
        return "-".ljust(52,'-')

    @staticmethod
    def addHeader(result):
        output = "\n"
        output += Style.MAGENTA + "*".ljust(REMAIN_START,'*') + "\n"
        output += Style.YELLOW
        output += " ___ ___ ___ _   _ _  _____  _ " + "\n"
        output += "| _ \ __/ __| | | | ||_   _|(_)" + "\n"
        output += "|   / _|\__ \ |_| | |__| |   _ " + "\n"
        output += "|_|_\___|___/\___/|____|_|  (_)" + "\n"
        output += " " + Style.MAGENTA + "\n"
        output += "*".ljust(REMAIN_START,'*') + "\n"
        return output

    @staticmethod
    def formatPrintTags(tags):
        output = ""
        if isinstance(tags,list):
           for tag in tags:
              output += Style.GREEN + tag["Key"] + Style.WHITE + "=" + Style.MAGENTA + tag["Value"] + Style.RESET + ", "
           return output   
        else:
           for tag in tags:
              output += Style.GREEN + str(tag) + Style.WHITE + "=" + Style.MAGENTA + str(tags[tag])+  Style.RESET + ", "
        if len(output)  > 2:    
           return output[:len(output)-2]    
        return Style.MAGENTA + "---" +  Style.RESET   

    @staticmethod
    def formatResult(functionName, result, config, jsonResult, tagsApply, tableArgs):
        # Show result at screen (verbose mode)
        if jsonResult != "" and (tableArgs.verbose or config.printResults):
           colorful_json = highlight(jsonResult, lexers.JsonLexer(), formatters.TerminalFormatter())
           print(" ")
           print(colorful_json)

        # Save to a file the JSON result
        if tableArgs and tableArgs.saveToFile and jsonResult != "":
           now = datetime.now()   
           fileName = now.strftime("%Y%m%d-%H%M%S") + "_" + functionName.replace("/","_").replace(".","_").replace(" ","-") + ".json"
           Utils.saveJsonToFile(fileName , jsonResult)             

        output = Utils.addHeader(result)
        output += Style.RESET + " " + Emoticons.pin() + " >>---> " + Style.GREEN + functionName +  Style.RESET + " <---<< "+ Emoticons.pin() + "\n"

        if config.awsTagsToFilter and tagsApply:
           output += "\n " + Emoticons.magnifier() + " Filters  \n      Environment Tags...: " + Style.GREEN + Utils.formatPrintTags(config.awsTags) + Style.RESET + "\n"
        if tableArgs and len(tableArgs.tagsTemp) > 0:
           output += "      Arguments   Tags...: " + Style.GREEN + Utils.formatPrintTags(tableArgs.tagsTemp) + Style.RESET + "\n"   

        output += " " + "\n"
        output += Style.RESET + result + "\n"
        output += " " + "\n"
        output += Style.MAGENTA + "*".ljust(REMAIN_START,'*') + "\n"
        output += " " + Style.RESET + "\n"
        return output

    @staticmethod 
    def dictToJson(response) :
        strObj = StringIO()
        json.dump(response, strObj, default=str, indent=4, sort_keys=True)
        return strObj.getvalue()

    @staticmethod
    def formatNumber(number, group, decimalPlaces, americanFormat=False):
        strNumber = str(number)
        if strNumber.find(".") > 1:
           intPart     = strNumber[:strNumber.find(".")]
           decimalPart = strNumber[strNumber.find(".")+1:]
        else:   
           intPart       = strNumber
           decimalPart   = ""
           decimalPlaces = 0

        groupSeparator   = "."
        decimalSeparator = ","
        if americanFormat:
           groupSeparator   = ","
           decimalSeparator = "."

        group     = 3
        strNumber = ""
        for c in reversed(intPart):
            strNumber = c + strNumber
            group -= 1
            if group == 0:
                strNumber = groupSeparator + strNumber
                group = 3
        
        if strNumber.startswith("."):
           strNumber = strNumber[1:]

        if decimalPlaces == 0 or decimalPart == "":
           return strNumber
        else:
           strNumber = strNumber + decimalSeparator
           for c in decimalPart:
               strNumber = strNumber + c
               decimalPlaces -= 1
               if decimalPlaces == 0:
                   break
           return strNumber

    @staticmethod
    def labelTimestamp():
        return datetime.now().strftime("%Y%m%d_%H%M%S")

    @staticmethod
    def clearScreen():
        if sys.platform == "linux" or sys.platform == "linux2":
           os.system('clear')
        elif sys.platform == "win32" or sys.platform == "win64":
           os.system('cls')   
        elif sys.platform == "darwin":
           os.system('clear')      
        else:
           os.system('clear')       

    # Try to copy a value to Clipboard
    @staticmethod
    def addToClipboard(text):
        try:
            textNoColor = Utils.removeCharsColors(text)
            if textNoColor:
               clipboard.copy(textNoColor.strip())
        except:
            pass      
        #
        # Option just for Ubuntu Linux (native)
        #try:
        #   command = 'echo "' + text + '" | xsel --clipboard'
        #   os.system(command)
        #except ValueError:
        #   # Ignore this error
        #   print(ValueError)

    @staticmethod
    def isNumber(s):
        try:
            int(s)
            return True
        except ValueError:
            return False   

            
    
    @staticmethod
    def removeCharsColors(text):
        if isinstance(text, str):
           text = text.replace(Style.UNDERLINE,"") \
                      .replace(Style.RESET,"") \
                      .replace(Style.BLACK,"") \
                      .replace(Style.RED,"") \
                      .replace(Style.GREEN,"") \
                      .replace(Style.YELLOW,"") \
                      .replace(Style.BLUE,"") \
                      .replace(Style.MAGENTA,"") \
                      .replace(Style.CYAN,"") \
                      .replace(Style.WHITE,"") \
                      .replace(Style.BBLACK,"") \
                      .replace(Style.BRED,"") \
                      .replace(Style.BGREEN,"") \
                      .replace(Style.BYELLOW,"") \
                      .replace(Style.BBLUE,"") \
                      .replace(Style.BMAGENTA,"") \
                      .replace(Style.BCYAN,"") \
                      .replace(Style.BWHITE,"") \
                      .replace(Style.BG_BLACK,"") \
                      .replace(Style.BG_RED,"") \
                      .replace(Style.BG_GREEN,"") \
                      .replace(Style.BG_YELLOW,"") \
                      .replace(Style.BG_BLUE,"") \
                      .replace(Style.BG_PURPLE,"") \
                      .replace(Style.BG_CYAN,"") \
                      .replace(Style.BG_WHITE,"") \
                      .replace(Style.IBLACK,"") \
                      .replace(Style.IRED,"") \
                      .replace(Style.IGREEN,"") \
                      .replace(Style.IYELLOW,"") \
                      .replace(Style.IBLUE,"") \
                      .replace(Style.IMAGENTA,"") \
                      .replace(Style.ICYAN,"") \
                      .replace(Style.IWWHITE,"") \
                      .replace(Style.BIBLACK,"") \
                      .replace(Style.BIRED,"") \
                      .replace(Style.BIGREEN,"") \
                      .replace(Style.BIYELLOW,"") \
                      .replace(Style.BIBLUE,"") \
                      .replace(Style.BIPURPLE,"") \
                      .replace(Style.BICYAN,"") \
                      .replace(Style.BIWHITE,"") \
                      .replace(Style.On_IBLACK,"") \
                      .replace(Style.On_IRED,"") \
                      .replace(Style.On_IGREEN,"") \
                      .replace(Style.On_IYELLOW,"") \
                      .replace(Style.On_IBLUE,"") \
                      .replace(Style.On_IPURPLE,"") \
                      .replace(Style.On_ICYAN,"") \
                      .replace(Style.On_IWHITE,"")
        return text


import boto3
import logging
import sys, os
import json
import math
import locale
import requests
from pygments import highlight, lexers, formatters
from botocore.exceptions import ClientError
from datetime import datetime
from datetime import timedelta  
from pyawscp.Functions import Functions
from pyawscp.Utils import Utils, Style
from pyawscp.PrettyTable import PrettyTable
from pyawscp.Emoticons import Emoticons
from pyawscp.Config import Config
from pyawscp.TableArgs import TableArgs
from pyawscp.pytransfers3.PyTransferS3 import PyTransferS3

LOG = logging.getLogger("app." + __name__)
LOG.setLevel(logging.INFO)

class PyS3CP:
    config = None

    def __init__(self, config):
        self.config = config

    def listBucketsS3(self):
        s3api      = self.botoSession().client('s3')
        csvFile    = None

        tableArgs  = TableArgs()
        tableArgs.setArguments(self.config.commandArguments) 

        if "," in self.config.commandArguments:
           tableArgs.setArguments(self.config.commandArguments)
           if ",csv" in self.config.commandArguments:
              csvFile = 'results/' + datetime.now().strftime("%Y%m%d-%H%M%S") + "_listBucketsS3.csv"
        else:    
           if "csv" in self.config.commandArguments:
              csvFile = 'results/' + datetime.now().strftime("%Y%m%d-%H%M%S") + "_listBucketsS3.csv"

        filters=[]
        if self.config.awsTagsToFilter():
           # Environment Tags set
           for tag in self.config.awsTags: 
               filters.append(
                   {'Name':'tag:' + tag,'Values':[self.config.awsTags[tag]]}
               )

        # Tags from command line arguments
        if len(tableArgs.tagsTemp) > 0:
           for tag in tableArgs.tagsTemp: 
               filters.append(
                   {'Name':'tag:' + tag,'Values':[tableArgs.tagsTemp[tag]]}
               )

        #Prefix="general/artifactory/"
        listBuckets = s3api.list_buckets()

        jsonResult = ""
        if self.config.printResults or tableArgs.verbose or tableArgs.saveToFile: 
           jsonResult = Utils.dictToJson(listBuckets)

        header = ["#","Name", "Creation Date"]
        prettyTable = PrettyTable(header)

        idx_lin    = 0
        totalCount = 0

        if csvFile:
           try:
               os.remove(csvFile)
           except OSError:
               pass

        #locale.setlocale(locale.LC_ALL, '')       

        for bucket in listBuckets["Buckets"]:

           idx_lin += 1

           bucketName   = bucket["Name"]
           creationDate = bucket["CreationDate"].strftime("%Y-%b-%d %H:%M")

           columns = [ str(idx_lin), bucketName, creationDate ]

           prettyTable.addRow(columns)

           #CVS File
           if csvFile:
              if not os.path.exists("results"):
                 os.makedirs("results")
              with open(csvFile, 'a') as the_file:
                 the_file.write(str(idx_lin) + "," + bucketName + "," + creationDate + "\n")

           totalCount += 1      


        resultTxt = ""
        resultTxt = " The List of S3 Buckets, total of..: " + Style.GREEN + format(totalCount,",") + Style.RESET
        
        if (int(tableArgs.sortCol) - 1) > len(columns):
            tableArgs.sortCol = "1"
            prettyTable.sortByColumn(int(tableArgs.sortCol) - 1)

        prettyTable.sortByColumn(0)   
        prettyTable.numberSeparator = True
        prettyTable.ascendingOrder(not tableArgs.desc)
        result = resultTxt + "\n\n" + prettyTable.printMe("listBucketsS3",self.config.tableLineSeparator, tableArgs)
        return Utils.formatResult(Functions.FUNCTIONS[Functions.LIST_BUCKETS_S3]["name"], result, self.config, jsonResult, True, tableArgs)    
    
    def listObjectsBucketS3(self):
        s3api      = self.botoSession().client('s3')
        bucketName = ""
        prefix     = ""
        csvFile    = None

        tableArgs  = TableArgs()
        tableArgs.setArguments(self.config.commandArguments) 

        if "," in self.config.commandArguments:
           bucketName = self.config.commandArguments.split(",")[0]
           tableArgs.setArguments(self.config.commandArguments)
           if ",csv" in self.config.commandArguments:
              csvFile = 'results/' + datetime.now().strftime("%Y%m%d-%H%M%S") + '_listObjectS3_' + bucketName.replace("/","_") + (prefix.replace("/","_") if prefix else "") + '.csv'           
        else:    
           bucketName = tableArgs.cleanPipelineArguments()

        if not bucketName:
           resultTxt = "Where is the bucket? You didn't tell me which Bucket " + Emoticons.ops() +  Style.RESET
           return Utils.formatResult(Functions.LIST_OBJECTS_BUCKETS_S3,resultTxt, self.config, "", True, tableArgs)

        if bucketName.find("/") > -1:
           prefix     = bucketName[bucketName.find("/")+1:]   
           bucketName = bucketName[:bucketName.find("/")]   

        filters=[]
        if self.config.awsTagsToFilter():
           # Environment Tags set
           for tag in self.config.awsTags: 
               filters.append(
                   {'Name':'tag:' + tag,'Values':[self.config.awsTags[tag]]}
               )

        # Tags from command line arguments
        if len(tableArgs.tagsTemp) > 0:
           for tag in tableArgs.tagsTemp: 
               filters.append(
                   {'Name':'tag:' + tag,'Values':[tableArgs.tagsTemp[tag]]}
               )

        #Prefix="general/artifactory/"
        bucketName = bucketName.replace("/","")
        prefix     = "" if prefix == "/" else prefix
        if prefix.startswith("/"):
           prefix = prefix[1:]
        try:   
            bucketObjects = s3api.list_objects_v2(Bucket=bucketName, Delimiter = "/", Prefix=prefix)
        except ClientError as ex:
            if ex.response['Error']['Code'] == "NoSuchBucket" or ex.response['Error']['Code'] == "AllAccessDisabled":
               resultTxt = Style.IRED + " Ops.. " + Emoticons.ops() + Style.RESET + " Most likely Bucket \"" + Style.GREEN + bucketName + prefix + Style.RESET + "\" does not exist!" +  Style.RESET
               return Utils.formatResult(Functions.LIST_OBJECTS_BUCKETS_S3,resultTxt, self.config, "", True, tableArgs)
            else:   
               resultTxt = Style.IRED + " Ops.. " + Emoticons.ops() + Style.RESET + " ERROR " + Style.GREEN + "\"" + ex.response['Error']['Code'] + "\"" + Style.RESET
               return Utils.formatResult(Functions.LIST_OBJECTS_BUCKETS_S3,resultTxt, self.config, "", True, tableArgs)
        else:
            if ("CommonPrefixes" not in bucketObjects) and ("Contents" not in bucketObjects): 
               resultTxt = " Ops.. " + Emoticons.ops() + " Nothing else to list inside \"" + Style.GREEN + bucketName + prefix + "\"" + Style.RESET
               return Utils.formatResult(Functions.LIST_OBJECTS_BUCKETS_S3,resultTxt, self.config, "", True, tableArgs)

            jsonResult = None
            if self.config.printResults or tableArgs.verbose or tableArgs.saveToFile: 
               jsonResult = {}
               jsonResult["s3Objects"] = []
            header = ["#","Object", "Count","Size (KB)","Last Modified", "Size (KB)", "Size (MB)", "Size (GB)","Storage Class"]
            prettyTable = PrettyTable(header)
            if csvFile:
               try:
                     os.remove(csvFile)
               except OSError:
                     pass
            #listObjectsBucketS3 onepay-us,prefix=logs/AWSLogs/659915611011/ | logs <-- List the content of each Bucket inside this one
            #listObjectsBucketS3 onepay-us,prefix=logs/AWSLogs/659915611011 | logs  <-- List the Current Bucket
            #locale.setlocale(locale.LC_ALL, '')       
            result = self._listObjectsBucketS3AppendCommonPrefixes(prettyTable, header, bucketObjects, csvFile, bucketName, prefix, s3api, jsonResult)
            totalSizeUnit = ""
            if result["totalSize"] > 0 and result["totalSize"] <= 1024:        
               totalSizeUnit = "Bytes"
            elif result["totalSize"] > 1024 and result["totalSize"] < 1000000:        
               result["totalSize"] = result["totalSize"] / 1000       
               totalSizeUnit = "KB"   
            elif result["totalSize"] >= 1000000 and result["totalSize"] < 1000000000:        
               result["totalSize"] = result["totalSize"] / 1000000       
               totalSizeUnit = "MB"
            else:   
               result["totalSize"] = result["totalSize"] / 1000000000        
               totalSizeUnit = "GB"
            
            resultTxt = ""
            resultTxt = " The S3 Bucket " + Style.GREEN + bucketName + "/" + prefix + Style.MAGENTA + " (not recursive)" + \
                        Style.RESET + "\n     Count of Objects.........: " + Style.GREEN + format(result["totalCount"],",") + \
                        Style.RESET + "\n     Size  of Objects.........: " + Style.GREEN + locale.format("%.2f",result["totalSize"], grouping=True) + " " + totalSizeUnit + Style.RESET
            
            #if (int(tableArgs.sortCol) - 1) > len(columns):
            #    tableArgs.sortCol = "1"
            #    prettyTable.sortByColumn(int(tableArgs.sortCol) - 1)
            prettyTable.sortByColumn(0)   
            prettyTable.numberSeparator = True
            prettyTable.ascendingOrder(not tableArgs.desc)
            result = resultTxt + "\n\n" + prettyTable.printMe("listObjectsBucketS3",self.config.tableLineSeparator, tableArgs)
            return Utils.formatResult(Functions.FUNCTIONS[Functions.LIST_OBJECTS_BUCKETS_S3]["name"], result, self.config, Utils.dictToJson(jsonResult), True, tableArgs)

    def _listObjectsBucketS3AppendCommonPrefixes(self, prettyTable, header, bucketObjects, csvFile, bucketName, prefix, s3api, jsonResult):
        result     = {}
        idx_lin    = 0
        totalSize  = 0
        totalCount = 0

        if "CommonPrefixes" in bucketObjects:
            for object in bucketObjects["CommonPrefixes"]:
                subFolder = object["Prefix"]
                #Count Contents subfolder previously ?!  or not!
                subFolderObjects    = s3api.list_objects_v2(Bucket=bucketName, Delimiter = "/", Prefix=subFolder)
                #print(subFolderObjects)
                subFolderTotalCount = 0
                subFolderTotalSize  = 0
                if "Contents" in subFolderObjects:
                   subFolderTotalCount = len(subFolderObjects["Contents"])
                   for objectSubfolder in subFolderObjects["Contents"]:
                       subFolderTotalSize += objectSubfolder["Size"]
                subFolderTotalSize = subFolderTotalSize / 1000       

                idx_lin += 1
                columns = [ str(idx_lin), subFolder, subFolderTotalCount, math.ceil(subFolderTotalSize) , "--" , "--", "--", "--", "--" ]
    
                if jsonResult:
                   jsonResult["s3Objects"].append({
                      "key": subFolder,
                      "countObjectsInside": subFolderTotalCount,
                      "sizeInKbObjectsInside": subFolderTotalSize,
                      "lastModified": "",
                      "sizeInKb": 0,
                      "sizeInMega": 0,
                      "sizeInGiga": 0
                   })   
    
                prettyTable.addRow(columns)
                if csvFile:
                   if not os.path.exists("results"):
                      os.makedirs("results")
                   with open(csvFile, 'a') as the_file:
                      the_file.write(str(idx_lin) + "," + subFolder + "," + "" +  ","+ "" + ","+ "" + "," + "" + "," + "" + "," + "" + "," + "" + "\n")   

        if "Contents" in bucketObjects:
            for object in bucketObjects["Contents"]:
                key = object["Key"]
                if True:
                #if key != prefix:
                   idx_lin += 1          
                   size        = object["Size"]
                   totalCount += 1
                   totalSize  += size
       
                   sizeInKb   = size / 1024
                   sizeInMega = size / 1000000
                   sizeInGiga = size / 1000000000       
       
                   columns = [ str(idx_lin), key, "--", "--", object["LastModified"].strftime("%d-%b-%y %H:%M"),
                               round(sizeInKb,2), round(sizeInMega,2), round(sizeInGiga,2), object["StorageClass"] ]        
       
                   if jsonResult:
                      jsonResult["s3Objects"].append({
                         "key": key,
                         "countObjectsInside": 0,
                         "sizeInKbObjectsInside": 0,
                         "lastModified": object["LastModified"].strftime("%d-%b-%y %H:%M"),
                         "sizeInKb": sizeInKb,
                         "sizeInMega": round(sizeInMega,2),
                         "sizeInGiga": round(sizeInGiga,2)
                      })            
                   
                   prettyTable.addRow(columns)
       
                   if csvFile:
                      if not os.path.exists("results"):
                         os.makedirs("results")
                      with open(csvFile, 'a') as the_file:
                         the_file.write(str(idx_lin) + "," + key + ",,," + object["LastModified"].strftime("%d-%b-%y %H:%M") + "," + str(sizeInKb)
                         + str(round(sizeInMega,2)) + "," + str(round(sizeInGiga,2)) + "," + object["StorageClass"] + "\n")

        result["idx_lin"]    = idx_lin
        result["totalSize"]  = totalSize
        result["totalCount"] = totalCount
        return result

    """
    # Replaced by the PyTransferS3 Function (same functionality)
    def listMultiPartUploads(self):
        s3api       = self.botoSession().client('s3')
        isRecursive = False
        bucketName  = ""
        prefix      = ""
        csvFile     = None

        tableArgs   = TableArgs()
        tableArgs.setArguments(self.config.commandArguments)

        if "," in self.config.commandArguments:
           bucketName = self.config.commandArguments.split(",")[0]
           tableArgs.setArguments(self.config.commandArguments)
           if ",recursive" in self.config.commandArguments:
              isRecursive = True
           if "prefix" in self.config.commandArguments:
              for arg in self.config.commandArguments.split(","):
                  if "prefix" in arg:
                      prefix = arg.split("=")[1]
           if ",csv" in self.config.commandArguments:
              csvFile = 'multiPartUpload_' + bucketName.replace("/","_") + (prefix.replace("/","_") if prefix else "") + '.csv'           
        else:    
           bucketName = tableArgs.cleanPipelineArguments()

        if not bucketName:
           resultTxt = "Where is the bucket? You didn't tell me which Bucket " + Emoticons.ops() +  Style.RESET
           return Utils.formatResult(Functions.LIST_MULTIPART_UPLOADS_S3,resultTxt, self.config, "", True, tableArgs)

        if bucketName.find("/") > -1:
           prefix     = bucketName[bucketName.find("/")+1:]   
           bucketName = bucketName[:bucketName.find("/")]    

        filters=[]
        if self.config.awsTagsToFilter():
           # Environment Tags set
           for tag in self.config.awsTags: 
               filters.append(
                   {'Name':'tag:' + tag,'Values':[self.config.awsTags[tag]]}
               )

        # Tags from command line arguments
        if len(tableArgs.tagsTemp) > 0:
           for tag in tableArgs.tagsTemp: 
               filters.append(
                   {'Name':'tag:' + tag,'Values':[tableArgs.tagsTemp[tag]]}
               )

        originalPrefix = prefix
        bucketName = bucketName.replace("/","")
        prefix     = "" if prefix == "/" else prefix
        if prefix.startswith("/"):
           prefix = prefix[1:]
        #bucketObjects  = s3api.list_multipart_uploads(Bucket=bucketName, Delimiter = "/", Prefix=prefix)
        bucketObjects  = s3api.list_multipart_uploads(Bucket=bucketName)

        jsonResult = ""
        if self.config.printResults or tableArgs.verbose or tableArgs.saveToFile:
           jsonResult = Utils.dictToJson(bucketObjects)

        #Debug
        #colorful_json = highlight(Utils.dictToJson(bucketObjects), lexers.JsonLexer(), formatters.TerminalFormatter())
        #print(colorful_json)

        header = ["#","Path","Key", "Initiated", "StorageClass", "Initiator ID", "Initiator Name"]
        prettyTable = PrettyTable(header)

        idx_lin = 0
        idxLin  = [idx_lin]
        if csvFile:
           try:
               os.remove(csvFile)
           except OSError:
               pass

        uploads = False
        columns = []
        if "Uploads" in bucketObjects:
            uploads = True
            path    = bucketObjects["Bucket"]
            for object in bucketObjects["Uploads"]:
                path          = bucketName + "/" + prefix
                key           = object["Key"]
                initiated     = object["Initiated"].strftime("%Y-%b-%d %H:%M") 
                storageClass  = object["StorageClass"]
                initiatorId   = object["Initiator"]["ID"]
                initiatorName = object["Initiator"]["DisplayName"]
                
                idx_lin += 1
                idxLin   = [idx_lin]
                columns  = [ str(idx_lin), path, key, initiated , storageClass, initiatorId, initiatorName ]        

                prettyTable.addRow(columns)

                if csvFile:
                   line = str(idx_lin) + "," + path + "," + key + "," + initiated + "," + storageClass + "," + initiatorId + "," + initiatorName + "\n"
                   #CVS File
                   with open(csvFile, 'a') as the_file:
                        the_file.write(line)
        
        if "CommonPrefixes" in bucketObjects and isRecursive:
            idxLin = [idx_lin]
            for children in bucketObjects["CommonPrefixes"]:
                prefix = children["Prefix"]
                self._listMultiPartUploadsChild(s3api,bucketName,prefix,idxLin,prettyTable,isRecursive,csvFile)

        resultTxt = ""
        resultTxt = " The S3 Bucket " + Style.GREEN + bucketName + "/" + originalPrefix + Style.MAGENTA + \
                    Style.RESET + "\n has total of " + Style.GREEN + format(idxLin[0],",") + \
                    Style.RESET + " multipart uploads not finished" + Style.RESET
        
        #if (int(tableArgs.sortCol) - 1) > len(columns):
        #    tableArgs.sortCol = "1"
        #    prettyTable.sortByColumn(int(tableArgs.sortCol) - 1)

        if uploads or len(prettyTable.content) > 1:
           prettyTable.sortByColumn(0)   
           prettyTable.numberSeparator = True
           prettyTable.ascendingOrder(not tableArgs.desc)
           result = resultTxt + "\n\n" + prettyTable.printMe("originName",self.config.tableLineSeparator, tableArgs)
           return Utils.formatResult(Functions.FUNCTIONS[Functions.LIST_MULTIPART_UPLOADS_S3]["name"], result, self.config,jsonResult, True, tableArgs)
        else:
           result = resultTxt + "\n\n" + Style.GREEN + " No Uploads found" + Style.RESET
           return Utils.formatResult(Functions.FUNCTIONS[Functions.LIST_MULTIPART_UPLOADS_S3]["name"], result, self.config,jsonResult, True, tableArgs)

    def _listMultiPartUploadsChild(self,s3api,bucketName,prefix,idxLin,prettyTable,isRecursive,csvFile):
        bucketObjects = s3api.list_multipart_uploads(Bucket=bucketName, Delimiter = "/", Prefix=prefix)

        #Debug
        #colorful_json = highlight(Utils.dictToJson(bucketObjects), lexers.JsonLexer(), formatters.TerminalFormatter())
        #print(colorful_json)

        if "Uploads" in bucketObjects:
            for object in bucketObjects["Uploads"]:
                path          = bucketName + "/" + prefix
                key           = object["Key"]
                initiated     = object["Initiated"].strftime("%Y-%b-%d %H:%M") 
                storageClass  = object["StorageClass"]
                initiatorId   = object["Initiator"]["ID"]
                initiatorName = object["Initiator"]["DisplayName"]

                idxLin[0] = idxLin[0] + 1
                columns  = [ str(idxLin[0]), path, key, initiated , storageClass, initiatorId, initiatorName ]        

                prettyTable.addRow(columns)

                if csvFile:
                   line = str(idxLin[0]) + "," + path + "," + key + "," + initiated + "," + storageClass + "," + initiatorId + "," + initiatorName + "\n"
                   #CVS File
                   with open(csvFile, 'a') as the_file:
                        the_file.write(line)

        if "CommonPrefixes" in bucketObjects and isRecursive:
            for children in bucketObjects["CommonPrefixes"]:
                prefix = children["Prefix"]
                self._listMultiPartUploadsChild(s3api,bucketName,prefix,idxLin,prettyTable,isRecursive,csvFile)
    """            

    # TODO: Add this as a function to Shell
    def preSignURL(self):
        tableArgs = TableArgs()
        tableArgs.setArguments(self.config.commandArguments)

        # one week
        #expiration = 604800
        # Default: One hour 
        expiration = 3600

        verbose = False
        args = tableArgs.cleanPipelineArguments()
        if not "," in args:
           print(Emoticons.ops() + " --> "+ Style.RED + "Ops... " + Style.RESET + " No arguments?!?\n       Check how to use it: " + Style.IBLUE + "help s3PreSignedURL\n")
           return None
        else:
           a = args.split(",")
           bucketName = a[0]
           objectName = a[1]
           if bucketName == "verbose" or objectName == "verbose":
              print(Emoticons.ops() + " --> "+ Style.RED + "Ops... " + Style.RESET + " The Bucket and Object must be the first and second paramater informed. Check it:" + Style.IBLUE + " help s3PreSignedURL\n")
              return None
           if len(a) > 2:
              if a[2] == "verbose":
                 verbose = True
              else:
                 try:
                    expiration = int(a[2])
                 except:
                    pass
           if len(a) > 3:
              if a[3] == "verbose":
                 verbose = True
              else:
                 try:
                    expiration = int(a[3])
                 except:
                    pass
        
        if expiration < 900 or expiration > 604800:
           print(Emoticons.ops() + " --> "+ Style.RED + "Ops... " + Style.RESET + " Expiration: minimum is " + Style.IGREEN + "15 minutes" + Style.RESET + ", maximum is " + Style.IGREEN + "7 days" + Style.RESET + ", please verify!" + Style.IBLUE + "   help s3PreSignedURL\n")
           return None
            
        s3api = self.botoSession().client('s3')
        # To Access the object
        url = s3api.generate_presigned_url('get_object',
                                             Params={
                                                'Bucket': bucketName,
                                                'Key': objectName
                                             },
                                             ExpiresIn=expiration,
                                             HttpMethod="GET")
        Utils.addToClipboard(url)
        if verbose:
           response = requests.get(url)
           print(Style.IGREEN + Emoticons.wink() + " --> Status")
           print(response)
           print(Style.IGREEN + Emoticons.wink() + " --> Object Content:")
           print(response.text)
        print(Style.IGREEN + Emoticons.wink() + " --> Ok!")
        print(Style.GREEN + "-".ljust(80,"-") + "\n" + Style.IBLUE + url + Style.IBLUE + "\n" + Style.GREEN + "-".ljust(80,"-"))
        print(Emoticons.wink() + " --> " + Style.IGREEN +"The " + Style.IBLUE + "URL" + Style.GREEN + " it copied to your clipboard already!\n\n")
        return url

    def ajustProfile(self):
        return "default" if not self.config.awsProfile or self.config.awsProfile == "" else self.config.awsProfile

    def listUploads(self):
        tableArgs  = TableArgs()
        tableArgs.setArguments(self.config.commandArguments) 

        args = tableArgs.cleanPipelineArguments()
        if "," in args:
           bucket  = args.split(",")[0]
        else:
           bucket  = args

        if not bucket or bucket == "":
           msg = "\n" + \
                 Style.IRED + Emoticons.error() + Style.GREEN + " Ops! First, you need to tell the name of the Bucket..." + \
                 Style.RESET + "\n"
           return msg
           

        prefix = None
        if "/" in bucket:
           prefix = bucket[bucket.find("/")+1:]    
           bucket = bucket[:bucket.find("/")]   
           if prefix[len(prefix)-1:] != "/":
              prefix = prefix + "/"

        pyTransferS3 = PyTransferS3(True,self.config,bucket,self.ajustProfile(),self.config.awsRegion,prefix)
        response     = pyTransferS3.listUploads()
        jsonResult   = response["jsonResult"]
  
        header      = ["#","Key", "Initiated Date","Upload Id"]
        prettyTable = PrettyTable(header)
        idx_lin     = 0

        if (response["foundBucket"]):
           if "Uploads" in jsonResult:
              for u in jsonResult["Uploads"]:
                    idx_lin += 1
                    columns = [ str(idx_lin), u["Key"], u["Initiated"].strftime("%Y-%m-%d %H:%M"),u["UploadId"] ]
                    prettyTable.addRow(columns)
                 
              prettyTable.sortByColumn(0)   
              prettyTable.numberSeparator = True
              prettyTable.ascendingOrder(not tableArgs.desc)
              resultTxt = " The S3 Bucket " + Style.GREEN + bucket + ("/" + prefix if prefix else "") + Style.MAGENTA + \
                          Style.RESET + "\n has total of " + Style.GREEN + format(response["total"],",") + \
                          Style.RESET + " multipart uploads not finished" + Style.RESET
              result = resultTxt + "\n\n" + prettyTable.printMe("listUploads",self.config.tableLineSeparator, tableArgs)                                     
           else:
              result = "\033[32mNo Uploads found for \033[35m{}\033[0m".format(bucket)
        else:
           result = response["result"]
          
        return Utils.formatResult(Functions.FUNCTIONS[Functions.LIST_UPLOADS_S3]["name"], result, self.config, Utils.dictToJson(jsonResult), False, tableArgs)

    def abortUploads(self):
        tableArgs  = TableArgs()
        tableArgs.setArguments(self.config.commandArguments) 

        args = tableArgs.cleanPipelineArguments()
        if "," in args:
           bucket  = args.split(",")[0]
        else:
           bucket  = args

        prefix = None
        if "/" in bucket:
           prefix = bucket[bucket.find("/")+1:]    
           bucket = bucket[:bucket.find("/")]   
           if prefix[len(prefix)-1:] != "/":
              prefix = prefix + "/"

        pyTransferS3 = PyTransferS3(True,self.config,bucket,self.ajustProfile(),self.config.awsRegion,prefix)
        response     = pyTransferS3.abortUploads()

        header       = ["#","Upload Id", "Key"]
        prettyTable  = PrettyTable(header)
        idx_lin      = 0
        if not response["found"]:
           result = "Bucket \033[32m{}\033[0m was not found!".format(response["bucket"])   
        elif not response["nothing"]:
           for u in response["uploadsAborted"]:
              idx_lin += 1
              columns = [ str(idx_lin), u["uploadId"], u["key"] ]
              prettyTable.addRow(columns)
   
           prettyTable.sortByColumn(0)   
           prettyTable.numberSeparator = True
           prettyTable.ascendingOrder(not tableArgs.desc)
           resultTxt = " The S3 Bucket " + Style.GREEN + bucket + ("/" + prefix if prefix else "") + Style.MAGENTA + \
                       Style.RESET + "\n had total of " + Style.GREEN + format(response["total"],",") + \
                       Style.RESET + " multipart uploads not finished, now \033[31maborted!\033[0m" + Style.RESET
           result = resultTxt + "\n\n" + prettyTable.printMe("abortUploads",self.config.tableLineSeparator, tableArgs)    
        else:
           result = "Nothing to abort at \033[32m{}\033[0m".format(response["bucket"])   
        return Utils.formatResult(Functions.FUNCTIONS[Functions.LIST_UPLOADS_S3]["name"], result, self.config, Utils.dictToJson(response), False, tableArgs)

    def transferToS3(self):
        tableArgs  = TableArgs()
        tableArgs.setArguments(self.config.commandArguments) 

        args = tableArgs.cleanPipelineArguments()
        if "," in args:
           bucket    = args.split(",")[0]
           path      = args.split(",")[1]
           recursive = True if len(args.split(",")) > 2 and args.split(",")[2] == "recursive" else False
        else:
           bucket  = args
           path    = None

        if not bucket:
           msg = "\n" + \
                 Style.IRED + Emoticons.error() + Style.GREEN + " Ops! First, you need to tell the bucket to copy to...\n" + \
                 " Ex: " + Style.BLUE + "\n      transferToS3 \033[95mupload-temp-logs\033[34m,C:\MyFiles\Logs-2015-08-28.tar.gz" + \
                 Style.RESET + "\n"
           return msg
        if not path:   
           msg = "\n" + \
                 Style.IRED + Emoticons.error() + Style.GREEN + " Ops! You also need to tell the File/Folder to copy to \033[95m{}\033[0m\n".format(bucket) + \
                 " Ex: " + Style.BLUE + "\n      transferToS3 {},\033[95mC:\MyFiles\Logs-2015-08-28.tar.gz\033[0m".format(bucket) + \
                 Style.RESET + "\n"
           return msg

        prefix = None
        if "/" in bucket:
           prefix = bucket[bucket.find("/")+1:]    
           bucket = bucket[:bucket.find("/")]   
           if prefix[len(prefix)-1:] != "/":
              prefix = prefix + "/"

        pyTransferS3        = PyTransferS3(True,self.config,bucket,self.ajustProfile(),self.config.awsRegion,prefix)
        pyTransferS3.target = path
        pyTransferS3.transfer(recursive)
        return ""

    def showTargetsS3(self):
        pyTransferS3 = PyTransferS3(True,self.config,"",self.ajustProfile(),self.config.awsRegion,None)
        pyTransferS3.showTargets()
    def showTargetUploadsS3(self):
        tableArgs  = TableArgs()
        tableArgs.setArguments(self.config.commandArguments) 
        args = tableArgs.cleanPipelineArguments()

        if args == "":
           msg = "\n" + \
                 Style.IRED + Emoticons.error() + Style.GREEN + " Missing target parameter..." + \
                 Style.RESET + "\n"
           print(msg)            
           return msg

        if "," in args:
           target    = args.split(",")[0]
           showAll   = True if args.split(",")[1].lower() == "showAll".lower() else False
        else:
           target  = args
           showAll = False

        pyTransferS3 = PyTransferS3(True,self.config,"",self.ajustProfile(),self.config.awsRegion,None)
        pyTransferS3.target = target
        pyTransferS3.initDb()
        pyTransferS3.showTargetUploads(showAll)

    def showTargetUploadPartsS3(self):
        tableArgs  = TableArgs()
        tableArgs.setArguments(self.config.commandArguments) 
        args = tableArgs.cleanPipelineArguments()

        if "," in args:
           target    = args.split(",")[0]
           key       = args.split(",")[1] 
        else:
           msg = "\n" + \
                 Style.IRED + Emoticons.error() + Style.GREEN + " Missing the key to show its uploded parts..." + \
                 Style.RESET + "\n"
           print(msg)            
           return msg

        pyTransferS3 = PyTransferS3(True,self.config,"",self.ajustProfile(),self.config.awsRegion,None)
        pyTransferS3.target = target
        pyTransferS3.initDb()
        pyTransferS3.showTargetUploadsParts(key)

    def removeTargetS3(self):
        tableArgs  = TableArgs()
        tableArgs.setArguments(self.config.commandArguments) 
        args = tableArgs.cleanPipelineArguments()

        if args == "":
           msg = "\n" + \
                 Style.IRED + Emoticons.error() + Style.GREEN + " Missing target parameter..." + \
                 Style.RESET + "\n"
           print(msg)      
           return msg
        target = args

        pyTransferS3 = PyTransferS3(True,self.config,"",self.ajustProfile(),self.config.awsRegion,None)
        pyTransferS3.target = target
        pyTransferS3.initDb()
        pyTransferS3.removeTargetHistory()

    def botoSession(self):
        return self.config.botoSession()

if __name__ == '__main__':
    config = Config()
    config.awsProfile   = "default"
    config.awsRegion    = "eu-west-3"
    #config.awsProfile   = "ecomm"
    #config.awsRegion    = "eu-central-1"
    config.printResults = True
    #config.awsTags["Environment"] = "production"
    config.tableLineSeparator = False
    
    # subnets, is-public (only with subnets active)
    # sort1,desc (when requested together with subnets, the group lines will be omitted)
    if len(sys.argv) > 1:
       config.commandArguments = sys.argv[1]
    #else:
    #   config.commandArguments = "emagin-delivery,prefix=general/oracle/"
    pyS3CP = PyS3CP(config)
    report = pyS3CP.preSignURL("pyawscp","PythonAWSCP_SimpleFunction_listVPC.mp4")
    if report: 
       Utils.addToClipboard(report)
       print(report)



import os

class Emoticons:
    _PROMPT      = ['⛅','⛅']
    _WAITING     = ['💤💤💤','☕☕☕']
    _SEE_YA      = ['👋','✌']
    _ERROR       = ['❌💣','❌☠']
    _TOOL        = ['🔧','⚒']
    _THUMBS_UP   = ['👍','✔']
    _POINT_RIGHT = ['👉','❖']
    _WINK        = ['😉','☻']
    _OPS         = ['😕','☹']
    _PIN         = ['📌','✎']
    _ENV         = ['📝','✍']
    _TIME        = ['🕘','☕']
    _WAIT_DISTR  = ['🍺','♨']
    _WAIT_DISTR2 = ['🍼','⚾']
    _MAGNIFIER   = ['🔍','☌']
    _BLOCKS      = ['📦','❒']
    _REDMARK     = ['🔴','⚫']
    _UPLOAD      = ['📤','✈']
    _UPLOAD_PART = ['🔹','➩']
    _FOLDER      = ['🔹','➩']
    _OK          = ['✅','✅']
    _COOKIE      = ['🍪','⚫']
    _IMGS        = [
                     ['🕒','🕓','🕔','🕕','🕖','🕗','🕘','🕙','🕚','🕛','🕐','🕑'],
                     ['☰','☱','☲','☴','☵','☶','☷','☶','☴']
                   ]

    @staticmethod
    def isWindows():
        return 1 if os.name == "nt" else 0
    
    @staticmethod
    def cookie():
        return Emoticons._COOKIE[Emoticons.isWindows()]
    @staticmethod
    def prompt():
        return Emoticons._PROMPT[Emoticons.isWindows()]
    @staticmethod
    def waiting():
        return Emoticons._WAITING[Emoticons.isWindows()]    
    @staticmethod
    def seeYa():
        return Emoticons._SEE_YA[Emoticons.isWindows()]
    @staticmethod
    def error():
        return Emoticons._ERROR[Emoticons.isWindows()]
    @staticmethod
    def tool():
        return Emoticons._TOOL[Emoticons.isWindows()]
    @staticmethod
    def thumbsUp():
        return Emoticons._THUMBS_UP[Emoticons.isWindows()] 
    @staticmethod
    def pointRight():
        return Emoticons._POINT_RIGHT[Emoticons.isWindows()]
    @staticmethod
    def wink():
        return Emoticons._WINK[Emoticons.isWindows()]
    @staticmethod
    def pin():
        return Emoticons._PIN[Emoticons.isWindows()]
    @staticmethod
    def env():
        return Emoticons._ENV[Emoticons.isWindows()]
    @staticmethod
    def time():
        return Emoticons._TIME[Emoticons.isWindows()]
    @staticmethod
    def waitDistract():
        return Emoticons._WAIT_DISTR[Emoticons.isWindows()]
    @staticmethod
    def waitDistract2():
        return Emoticons._WAIT_DISTR2[Emoticons.isWindows()]
    @staticmethod
    def ops():
        return Emoticons._OPS[Emoticons.isWindows()]
    @staticmethod
    def magnifier():
        return Emoticons._MAGNIFIER[Emoticons.isWindows()]
    @staticmethod
    def ok():
        return Emoticons._OK[Emoticons.isWindows()]    
    @staticmethod
    def redMark():
        return Emoticons._REDMARK[Emoticons.isWindows()]        


import webbrowser
import os, sys
import threading
import pathlib
import logging

from pyawscp.pymxgraph.HttpServer import HttpServer
sys.path.append('..')
from pyawscp.Utils import Utils
from pyawscp.Emoticons import Emoticons

LOG = logging.getLogger("app." + __name__)
LOG.setLevel(logging.INFO)

class PymxGraphAwsNavigator:

    def __init__(self, pathToResources, images, htmlSnippets, verbose=False):
        self.pathToResources = pathToResources
        self.images          = images
        self.htmlSnippets    = htmlSnippets
        self.verbose         = verbose
        self.httpServer      = None

    def startHttpServer(self):
        self.httpServer = HttpServer(self.verbose)
        self.httpServer.run()

    def stopHttpServer(self):
        self.httpServer.stop()

    def openHTML(self,html):
        directory = os.path.dirname(os.path.realpath(__file__))
        fileName  = os.path.join(str(directory),"aws-navigator.html")
        
        try:
            os.remove(fileName)
        except OSError:
            pass

        with open(fileName, "w") as file:
            file.write(html)
        webbrowser.open('file://' + os.path.realpath(fileName))   

    def drawAwsNavigator(self):
        html = """
        <html>
          <head>
             """ + self.htmlSnippets.htmlTitle('Hi There!') + """
             """ + self.htmlSnippets.jsLibraries()          + """
             <style type="text/css">
                .modalInfo {
                    position: absolute;
                    display: block;
                    top: 200px;
                    left: 200px;
                    background: rgb(245, 245, 245);
                    //width: 350px;
                    //height: 250px;
                    padding: 5px;
                    font-size: 0.875rem;
                    box-shadow: 0 0 2.5rem rgba(0, 0, 0, 0.5);
                    border-radius: 0.3125rem;
                    font-family: 'Source Sans Pro', sans-serif
                }
                .content {
                    font-size: 0.8rem;
                    font-family: 'Consolas', sans-serif
                }
                .tag {
                    border: 1.4px solid darkcyan;
                    padding: 3px;
                    border-radius: 8px;
                    background-color:rgb(0,220,220);
                    color:black;
                    display:inline;
                    font-size: 12px;
                }
                .tableInfo {
                    font-family: Consolas;
                    font-size: 13px;
                    border: 2px solid darkcyan;
                    padding: 0;
                    border-radius: 8px;
                }
                .cellHeader {
                    border-bottom: 1.5px darkcyan solid;
                    font-size: 15px;
                    font-weight: bold;
                    background-color: darkcyan;
                }
                .cellInfo {
                    padding: 10px 8px 10px 8px;
                }
                .cellTags {
                    border-top: 1.5px darkcyan solid;
                    padding: 8px 8px 15px 8px;
                }
                .buttonCopyClipboard {
                    margin-bottom: -3px;
                }
             </style>
             <script type="text/javascript">

                 """ + self.htmlSnippets.jsDataURIForImage()    + """
                 
                 var requestId   = 0;

                 const Http             = new XMLHttpRequest();
                 const url              = '"""+ self.htmlSnippets.serverURL +"""';
                 var   previousGraphXml = [];
                 var   previousCoord    = {"x":0, "y":0};

                 function moveMe(graph, cellId, valueX, valueY, addToIt) {
                    if ( graph.view.getScale() == 1 ) { 
                        try
                        {
                            var v1 = graph.getModel().getCell(cellId); 
                            graph.clearSelection();
                            graph.getModel().beginUpdate(); 
                            var geo = graph.getCellGeometry(v1);
                            geo  = geo.clone();

                            previousCoord.x = geo.x;
                            previousCoord.y = geo.y;
                            
                            if ( addToIt ) {
                                geo.x += valueX;
                                geo.y += valueY;
                            } else {
                                geo.x = valueX;
                                geo.y = valueY;
                            }
                            graph.getModel().setGeometry(v1, geo);
                        }
                        finally
                        {
                            var morph = new mxMorphing(graph, 0, 0, 0);
                            morph.addListener(mxEvent.DONE, function()
                            {
                                graph.getModel().endUpdate();
                            });
                            morph.startAnimation();
                            return previousCoord;
                        }
                    }
                 }

                 function historyGraphGoBack(graph) {
                    // Clean any ModalInfo Timer started before go back 
                    if ( timerShowModalSet != undefined ) {
                            clearTimeout(timerShowModalSet)
                    }
                    level           -= 1
                    var parentCellId = '';
                    try {
                        graph.getModel().beginUpdate();
                        graph.removeCells(graph.getChildVertices(graph.getDefaultParent()));
                        var xml = previousGraphXml.pop();
                        var doc = mxUtils.parseXml(xml);
                        var codec = new mxCodec(doc);
                        var model = codec.decode(doc.documentElement);

                        // Find out the parent of the previous level to highlight it with some animation. 
                        // Example: if we go back to List_VPC level, we need to find the first VPC Cell and use its ParecenCellId
                        for (var key in model.cells)
                        {
                            var tmp = model.getCell(key);
                            if ( tmp.data != undefined && tmp.data.parentCellId != undefined ) {
                                if ( LEVELS[level] == LIST_VPC && tmp.data.awsResourceType == "vpc" ) {
                                    parentCellId = tmp.data.parentCellId;
                                    break;
                                } else
                                if ( LEVELS[level] == LIST_SUBNET && tmp.data.awsResourceType == "subnet" ) {
                                    parentCellId = tmp.data.parentCellId;
                                    break;
                                }
                            }
                        }
                        graph.getModel().mergeChildren(model.getRoot().getChildAt(0), graph.getDefaultParent());
                    } finally {
                        graph.getModel().endUpdate();
                    }
                    // Animation
                    var xs = [10,-10,-5,5,20,25];
                    var ys = [10,-10,-5,5,20,25];
                    var x  = xs[Math.floor(Math.random() * (5 - 0) + 0)];
                    var y  = ys[Math.floor(Math.random() * (5 - 0) + 0)];
                    var previousCoord;
                    setTimeout(function(){
                        previousCoord = moveMe(graph,parentCellId,x,y,true);
                    },70);
                    setTimeout(function(){
                        moveMe(graph,parentCellId,previousCoord.x,previousCoord.y, false);
                    },450);
                 }

                 function showModalInfo(currentState) {
                    // Check if the Mouse is over the same Object/Icon for a specific amount of time (set it on the calling of this throuhg setTimeOut, down below) 
                    // If the mouse has moved away after some time, the ModalInfo should not be shown, otherwise should be shown
                    var sameRegionYet = (globalCurrentState == currentState) && (globalCurrentState != undefined) && (globalCurrentState.cell != undefined);
                    if (!sameRegionYet)
                        return;

                    var divContent = document.getElementById(divContentId);
                    divContent.innerHTML  = currentState.cell.modalInfo.html;

                    if ( !showModalRightBtnFixed )
                         document.querySelector("#imgBtnClose").style.display = "none";

                    x = currentState.cell.getGeometry().x + (currentState.cell.getGeometry().width);
                    y = currentState.cell.getGeometry().y - 50;
                    var div = document.getElementById(divModalInfoId);
                    
                    var win = window,
                        doc = document,
                        docElem = doc.documentElement,
                        body    = doc.getElementsByTagName('body')[0],
                        width   = win.innerWidth || docElem.clientWidth || body.clientWidth,
                        height  = win.innerHeight|| docElem.clientHeight|| body.clientHeight;

                    div.style.display = "block";    
                        
                    // Adjust the Y position of the Div accordingly with the screen height    		 
                    divModalInfoHeight = div.offsetHeight;
                    if ( (y + divModalInfoHeight) > height ) {
                        y = y - ((y + divModalInfoHeight) - height) - (currentState.cell.getGeometry().height * 1.8); 
                    }
                    if ( y < 0 ) {
                        y = currentState.cell.getGeometry().y + 30;
                    }

                    // Adjust the X position of the Div accordingly with the screen width and the cell width
                    divModalInfoWidth = div.offsetWidth;
                    if ( (x + divModalInfoWidth) > width ) {
                        x = x - (divModalInfoWidth + currentState.cell.getGeometry().width + 5); 
                    }
                    
                    div.style.display = "block";
                    div.style.left    = x;
                    div.style.top     = y;
                 }

                 var delayTimeShowModalInfo = 500; // one second
                 var timerShowModalSet      = undefined;
                 var divModalInfoWidth      = 350;
                 var divModalInfoHeight     = 250;
                 var globalCurrentState     = null;
                 var showModalRightBtnFixed = false;
                 var divModalInfo           = document.createElement("div");   
                 var divModalInfoId         = "divModalInfo"
                 divModalInfo.style.display = "none";
                 divModalInfo.className     = "modalInfo";
                 //divModalInfo.innerHTML     = "Header";
                 divModalInfo.id            = divModalInfoId;
                 var divContentId           = "divContentId";
                 var divContent             = document.createElement("div");
                 divContent.innerHTML       = "Here, you can put any additional information";
                 divContent.id              = divContentId;
                 divModalInfo.appendChild(divContent);    

                 function main(container)
                 {
                     document.body.appendChild(divModalInfo);   

                     """ + self.htmlSnippets.jsBrowserNotSupported() + """
                     else
                     {
                         // Speedup the animation
                         mxText.prototype.enableBoundingBox = false
                         mxEvent.disableContextMenu(document.body);

                         // Creates the graph inside the given container
                         var graph = new mxGraph(container);
                         
                         // Disables all built-in interactions
                         graph.setEnabled(false);
                         graph.setPanning(true);
                         graph.panningHandler.useLeftButtonForPanning = true;
                         graph.setHtmlLabels(true);
         
                         // Handles clicks on cells
                         graph.addListener(mxEvent.CLICK, function(sender, evt)
                         {
                             var rightButton;
                             if ( evt.properties != undefined && evt.properties.event != undefined && evt.properties.event.button != undefined ) {
                                 rightButton = evt.properties.event.button == 2 ? true : false;
                             }
                             if ( rightButton != undefined && rightButton == false ) {
                                  var cell = evt.getProperty('cell');
                                  if (cell != null)
                                  {
                                      // If the Cell clicked is a VPC Resource and we are at the level of LIST_SUBNET, so...
                                      // we are clicking at the center of the graph, that is, the root parent (VPC) of all Subnets 
                                      // Then, we must go back to the previous Level: LIST_VPC(1) ---> AWS (0)
                                      if ( cell.data != undefined && cell.data.awsResourceType == "vpc" && LEVELS[level] == LIST_SUBNET ) {
                                          historyGraphGoBack(graph);
                                      } else
                                      if ( cell.data != undefined && cell.data.awsResourceType == "subnet" && LEVELS[level] == LIST_EC2 ) {
                                          historyGraphGoBack(graph);
                                      } else {
                                          // Before going foward, clean any ModalInfo TimeOut Set
                                          if ( timerShowModalSet != undefined ) {
                                              clearTimeout(timerShowModalSet)
                                          }
                                          // Save snapshot of the previous graph to go back when requested
                                          var enc          = new mxCodec();
                                          var node         = enc.encode(graph.getModel());
                                          var tmpXml       = mxUtils.getXml(node);
                                          if ( callAWSGraphNavigation(graph, cell) ) {
                                              // Only save when any call was implemented
                                              previousGraphXml.push(tmpXml);    
                                          }
                                      }
                                  }
                             } else {
                                 // Right Mouse Button Pressed
                                 showModalRightBtnFixed = true;
                                 showModalInfo(globalCurrentState);
                             }
                         });

                         // Handles mouseOver, mouseOut - ModalInfo
                         graph.addMouseListener({
                            currentState: null,
                            mouseMove: function(sender, evt) {
                                 if ( !showModalRightBtnFixed ) {
                                    if (this.currentState != null && (evt.getState() == this.currentState || evt.getState() == null) ) {
                                        globalCurrentState = this.currentState;
                                        if ( timerShowModalSet != undefined ) {
                                            clearTimeout(timerShowModalSet)
                                        }
                                        timerShowModalSet = setTimeout(showModalInfo,delayTimeShowModalInfo,this.currentState);
                                        if ( evt.getState() ) {
                                            evt.getState().setCursor('pointer');
                                        }     
                                    }

                                    var tmp = graph.view.getState(evt.getCell()); 
                                    
                                    if ( tmp != this.currentState ) {
                                        var div = document.getElementById(divModalInfoId);
                                        div.style.display = "none";
                                        if ( timerShowModalSet != undefined ) {
                                            clearTimeout(timerShowModalSet)
                                        }
                                        if ( tmp != null && tmp.cell.isVertex() ) {
                                            this.currentState  = tmp;
                                            globalCurrentState = tmp;
                                        }
                                    }
                                 }
                            },
                            mouseDown: function(sender, evt) { },
                            mouseUp:   function(sender, evt) { },
                            dragEnter: function(evt, state) { },
                            dragLeave: function(evt, state) { }
                         });

                         // Mouse Wheel Zooming
                         mxEvent.addMouseWheelListener((evt, up) => {
                            if (mxEvent.isConsumed(evt)) {
                                return;
                            }
                            let gridEnabled = graph.gridEnabled;
                            // disable snapping
                            graph.gridEnabled = false;
                            let p1 = graph.getPointForEvent(evt, false);
                            if (up) {
                                graph.zoomIn();
                            } else {
                                graph.zoomOut();
                            }
                            let p2 = graph.getPointForEvent(evt, false);
                            let deltaX = p2.x - p1.x;
                            let deltaY = p2.y - p1.y;
                            let view = graph.view;
                            view.setTranslate(view.translate.x + deltaX, view.translate.y + deltaY);
                            graph.gridEnabled = gridEnabled;
                            mxEvent.consume(evt);
                         }, container);
                          
                         // Changes the default vertex style in-place
                         var style = graph.getStylesheet().getDefaultVertexStyle();
                         // Rectangle (rounded)
                         style[mxConstants.STYLE_SHAPE] = mxConstants.SHAPE_RECTANGLE;
                         style[mxConstants.STYLE_ROUNDED] = true;
                         style[mxConstants.STYLE_PERIMETER] = mxPerimeter.RectanglePerimeter;
                         style[mxConstants.STYLE_GRADIENTCOLOR] = 'white';
                                         
                         // Gets the default parent for inserting new cells. This
                         // is normally the first child of the root (ie. layer 0).
                         var parent = graph.getDefaultParent();
         
                         var cx = graph.container.clientWidth / 2;
                         var cy = graph.container.clientHeight / 2;
                         
                         // Root of All
                         var awsImg = dataURIFor("aws-4.png").replace(";base64,",",");
                         var prefix = 'shape=image;image=' + awsImg + ";";

                         var cell       = graph.insertVertex(parent, 'AWS', 'AWS', cx - 20, cy - 15, 60, 40,prefix+'verticalLabelPosition=bottom;verticalAlign=top');
                         cell.data      = new CustomData('aws', 'root');
                         cell.modalInfo = new ModalInfoHTML("root",{"aws":"aws"});
                         
                         // Animates the changes in the graph model
                         graph.getModel().addListener(mxEvent.CHANGE, function(sender, evt)
                         {
                             var changes = evt.getProperty('edit').changes;
                             mxEffects.animateChanges(graph, changes);
                         });

                         """ +  self.htmlSnippets.jsMxToolBox("90","15") + """

                         var parameters = undefined;
                         postRequest(LIST_VPC,parameters,function(response) {
                             drillDown(graph, cell, response);
                         });
                     }
                 };

                 function closeModalInfo() {
                    var div = document.getElementById(divModalInfoId);
                    div.style.display      = "none";
                    showModalRightBtnFixed = false;
                 }

                 var LIST_VPC            = "listVpc";
                 var LIST_SUBNET         = "listSubnet";
                 var LIST_EC2            = "listEc2";
                 var LIST_SECURITY_GROUP = "listSecurityGroup";
                 var level               = 0;
                 var LEVELS              = ["ROOT", LIST_VPC, LIST_SUBNET, LIST_EC2];

                 function showLoading(show) {
                     if (show) {
                         document.querySelector("#loadingBackground").style.display = "";
                         document.querySelector("#loadingMessage").style.display = "";
                     } else {
                         document.querySelector("#loadingBackground").style.display = "none";
                         document.querySelector("#loadingMessage").style.display    = "none";
                     }
                 }

                 function callAWS(request, callback, event) {
                    showLoading(true);
                    postRequest(request.call,request.parameters,function(response, event) {
                        showLoading(false);
                        callback(response, event);
                    }, event);
                 }
                 function callListSecurityGroups(groupId, callback, event) {
                    var parameters  = [];
                    parameters.push({"name":"groupId","value":groupId});
                    var request = {
                        "call":LIST_SECURITY_GROUP,
                        "parameters": parameters
                    };
                    callAWS(request, callback, event);
                 }

                 function callAWSGraphNavigation(graph, cell) {
                     var callImplemented = false;
                     var awsResourceType = cell.data.awsResourceType;
                     if ( awsResourceType == "subnet" ) {
                         // Drill Down Subnets --> EC2s
                         showLoading(true);
                         callImplemented = true;
                         var parameters  = [];
                         parameters.push({"name":"vpcId","value":cell.data.awsResourceId},
                                         {"name":"subnetId","value":cell.data.awsResourceId});
                         postRequest(LIST_EC2,parameters,function(response) {
                             drillDown(graph, cell, response);
                             level = 3;
                         });
                     } else
                     if ( awsResourceType == "vpc" ) {
                         // Drill Down VPC --> Subnets
                         showLoading(true);
                         callImplemented = true;
                         var parameters  = [];
                         parameters.push({"name":"vpcId","value":cell.data.awsResourceId});
                         postRequest(LIST_SUBNET,parameters,function(response) {
                             drillDown(graph, cell, response);
                             level = 2;
                         });
                     } else
                     if ( awsResourceType == "aws" ) {
                         // Drill Down AWS --> VPCs
                         showLoading(true);
                         callImplemented = true;
                         var parameters  = undefined;
                         postRequest(LIST_VPC,parameters,function(response) {
                             drillDown(graph, cell, response);
                             level = 1;
                         });
                     }
                     return callImplemented;
                 }

                 function postRequest(action, parameters, callback, event) {
                     Http.open("POST", url, true);
                     Http.setRequestHeader('Content-type','application/json');
                     request = createRequest(action,parameters);
                     Http.onreadystatechange = (e) => {
                        if ( Http.readyState == 4 && Http.status == 200 ) {
                            callback(JSON.parse(Http.responseText), event)
                        } else {
                            //console.log("onreadystatechange = " + Http.status);
                        }
                     }
                     Http.onerror = (e) => {
                         showLoading(false);
                         alert("No connection available for " + url);
                     }
                     //Http.send(JSON.stringify(request));
                     Http.send(request);
                 }

                 function createRequest(action, params) {
                    var request        = {};
                    var parameters     = [];
                    request.action     = action;
                    request.parameters = parameters;

                    if ( params && params != undefined && params.length > 0 ) {
                        for (index = 0; index < params.length; ++index) {
                            request.parameters.push({
                                "name": params[index]["name"],
                                "value": params[index]["value"]
                            });
                        }
                    }
                    return JSON.stringify(request);
                }

                 // Loads the links for the given cell into the given graph
                 // by requesting the respective data in the server-side
                 // (implemented for this demo using the server-function)
                 function drillDown(graph, cell, json)
                 {
                     if (graph.getModel().isVertex(cell))
                     {
                         centerCell = cell;
                         var cx = graph.container.clientWidth / 2;
                         var cy = graph.container.clientHeight / 2;
                         
                         // Gets the default parent for inserting new cells. This
                         // is normally the first child of the root (ie. layer 0).
                         var parent = graph.getDefaultParent();
         
                         // Adds cells to the model in a single step
                         graph.getModel().beginUpdate();
                         try
                         {
                             var xml = jsonToGraphModel(cell.id, json);
                             var doc = mxUtils.parseXml(xml);
                             var dec = new mxCodec(doc);
                             var model = dec.decode(doc.documentElement);
         
                             // Removes all cells which are not in the response
                             for (var key in graph.getModel().cells)
                             {
                                 var tmp = graph.getModel().getCell(key);
                                 
                                 if (tmp != cell &&
                                     graph.getModel().isVertex(tmp))
                                 {
                                     graph.removeCells([tmp]);
                                 }
                             }
         
                             // Merges the response model with the client model
                             graph.getModel().mergeChildren(model.getRoot().getChildAt(0), parent);
         
                             // Moves the given cell to the center
                             var geo = graph.getModel().getGeometry(cell);
         
                             if (geo != null)
                             {
                                 geo = geo.clone();
                                 geo.x = cx - geo.width / 2;
                                 geo.y = cy - geo.height / 2;
         
                                 graph.getModel().setGeometry(cell, geo);
                             }
         
                             // Creates a list of the new vertices, if there is more
                             // than the center vertex which might have existed
                             // previously, then this needs to be changed to analyze
                             // the target model before calling mergeChildren above
                             var vertices = [];
                             
                             for (var key in graph.getModel().cells)
                             {
                                 var tmp = graph.getModel().getCell(key);
                                 
                                 if (tmp != cell && model.isVertex(tmp))
                                 {
                                     vertices.push(tmp);
         
                                     // Changes the initial location "in-place"
                                     // to get a nice animation effect from the
                                     // center to the radius of the circle
                                     var geo = model.getGeometry(tmp);
         
                                     if (geo != null)
                                     {
                                         geo.x = cx - geo.width / 2;
                                         geo.y = cy - geo.height / 2;
                                     }
                                 }
                             }
                             
                             // Arranges the response in a circle
                             var cellCount = vertices.length;
                             var phi = 2 * Math.PI / cellCount;
                             var r = Math.min(graph.container.clientWidth / 2.3,
                                              graph.container.clientHeight / 2.3);
                             
                             for (var i = 0; i < cellCount; i++)
                             {
                                 var geo = graph.getModel().getGeometry(vertices[i]);
                                 
                                 if (geo != null)
                                 {
                                     geo = geo.clone();
                                     geo.x += r * Math.sin(i * phi);
                                     geo.y += r * Math.cos(i * phi);
         
                                     graph.getModel().setGeometry(vertices[i], geo);
                                 }
                             }
                         }
                         finally
                         {
                             // Updates the display
                             graph.getModel().endUpdate();
                         }
                     }
                     showLoading(false);
                 };

                 
                 var colors      = ["rgb(0,0,255)","rgb(183,151,43)","rgb(110,105,89)","rgb(91,154,176)","rgb(255,0,9)","rgb(204,204,0)","rgb(128,128,128)","rgb(246,0,0)","rgb(246,178,103)","rgb(8,180,151)"];
                 var labelColors = {};
                 function colorForLabel(label) {
                    var color; 
                    if (labelColors[label] && labelColors[label] != undefined && labelColors[label] != "") {
                        color = labelColors[label];
                    }
                    if ( !color || color == undefined || color == "" ) {
                        var rnd = Math.floor(Math.random() * colors.length-1);
                        if ( rnd < 0 ) {
                            rnd = 3;
                        }
                        color = colors[rnd];
                        labelColors[label] = color;
                    }
                    return color;
                 }
                 function componentToHex(c) {
                    var hex = c.toString(16);
                    return hex.length == 1 ? "0" + hex : hex;
                 }
                 function rgbToHex(r, g, b) {
                    return "#" + componentToHex(r) + componentToHex(g) + componentToHex(b);
                 }
                 function rgbToHexStr(str) {
                    var cleanRGB = (str.replace("rgb(","").replace(")","")).split(","); 
                    return rgbToHex(parseInt(cleanRGB[0]), parseInt(cleanRGB[1]), parseInt(cleanRGB[2]));
                 }

                 function jsonToGraphModel(cellId, json)
                 {
                     // Creates a local graph with no display
                     var graph = new mxGraph();
                     
                     // Gets the default parent for inserting new cells. This
                     // is normally the first child of the root (ie. layer 0).
                     var parent = graph.getDefaultParent();
         
                     // Adds cells to the model in a single step
                     graph.getModel().beginUpdate();
                     try
                     {
                        if ( json != undefined ) {
                            action = json["aws"]["action"]

                            width  = 130;
                            height = 50;
                            
                            if (action == LIST_VPC) {
                                var v0  = graph.insertVertex(parent, cellId, 'AWS-Dummy', 0, 0, width, height);
                                v0.data    = new CustomData('root', undefined, json, '');
                                var prefix = 'shape=image;image=' + dataURIFor(AWS_VPC_IMG) + ";";
                                for( var key in json["aws"]["vpc"] ) {
                                    var vpc     = json["aws"]["vpc"][key];
                                    var id      = vpc.vpcId;
                                    var label   = "<b>" + vpc.vpcId + "</b><br>" + vpc.cidrBlock; 
                                    var v       = graph.insertVertex(parent, id, label, 0, 0, width, height, prefix+'verticalLabelPosition=bottom;verticalAlign=top;whiteSpace=wrap;');
                                    v.data      = new CustomData('vpc', vpc.vpcId, json, cellId);
                                    v.modalInfo = new ModalInfoHTML('vpc',vpc);
                                    var e       = graph.insertEdge(parent, null, '', v0, v);
                                }
                            } else
                            if (action == LIST_SUBNET) {
                                 var v0      = graph.insertVertex(parent, cellId, 'Vpc', 0, 0, width, height);
                                 v0.data     = new CustomData('vpc', undefined, json);
                                 var prefix1 = 'shape=image;image=' + dataURIFor(AWS_SUBNET_PRIVATE_IMG) + ";";
                                 var prefix2 = 'shape=image;image=' + dataURIFor(AWS_SUBNET_PUBLIC_IMG) + ";";
                                 var prefix  = prefix1;
                                 for( var key in json["aws"]["vpc"]["subnets"] ) {
                                    var subnet = json["aws"]["vpc"]["subnets"][key];
                                    var id           = subnet.subnetId;
                                    var defaultForAz = subnet.defaultForAz ? "<br>DEFAULT" : "";
                                    if (subnet.scope == "public") {
                                        prefix = prefix2
                                    } else {
                                        prefix = prefix1
                                    }
                                    var colorAz = colorForLabel(subnet.availabilitZone);
                                    var label   = "<b>" + subnet.subnetId + "</b><br>" + subnet.cidrBlock + "<br><span style='color:" + colorAz + ";'>" + subnet.availabilitZone + "</span>" + defaultForAz;
                                    var v       = graph.insertVertex(parent, id, label, 0, 0, width, height, prefix+'verticalLabelPosition=bottom;verticalAlign=top;whiteSpace=wrap;');
                                    v.data      = new CustomData('subnet', subnet.subnetId, json, cellId);
                                    v.modalInfo = new ModalInfoHTML('subnet',subnet);
                                    var e       = graph.insertEdge(parent, null,subnet.availabilitZone, v0, v,'strokeColor=' + rgbToHexStr(colorAz) + ';');
                                }
                            } else
                            if (action == LIST_EC2) {
                                 var v0     = graph.insertVertex(parent, cellId, 'Subnet', 0, 0, width, height);
                                 v0.data    = new CustomData('ec2', undefined, json);
                                 for( var key in json["aws"]["ec2"]["instances"] ) {
                                    var ec2instance = json["aws"]["ec2"]["instances"][key];

                                    var prefix = 'shape=image;image=' + dataURIFor(AWS_EC2_IMG) + ";";
                                    if ( ec2instance.instanceType.startsWith("t2") ) {
                                        prefix = 'shape=image;image=' + dataURIFor(AWS_EC2_T2_IMG) + ";";
                                    } else 
                                    if ( ec2instance.instanceType.startsWith("t3a") ) {
                                        prefix = 'shape=image;image=' + dataURIFor(AWS_EC2_T3A_IMG) + ";";
                                    } else
                                    if ( ec2instance.instanceType.startsWith("t3") ) {
                                        prefix = 'shape=image;image=' + dataURIFor(AWS_EC2_T3_IMG) + ";";
                                    } else
                                    if ( ec2instance.instanceType.startsWith("a1") ) {
                                        prefix = 'shape=image;image=' + dataURIFor(AWS_EC2_A1_IMG) + ";";
                                    } else
                                    if ( ec2instance.instanceType.startsWith("c4") ) {
                                        prefix = 'shape=image;image=' + dataURIFor(AWS_EC2_C4_IMG) + ";";
                                    } else
                                    if ( ec2instance.instanceType.startsWith("c5") ) {
                                        prefix = 'shape=image;image=' + dataURIFor(AWS_EC2_C5_IMG) + ";";
                                    } else
                                    if ( ec2instance.instanceType.startsWith("m4") ) {
                                        prefix = 'shape=image;image=' + dataURIFor(AWS_EC2_M4_IMG) + ";";
                                    } else
                                    if ( ec2instance.instanceType.startsWith("m5a") ) {
                                        prefix = 'shape=image;image=' + dataURIFor(AWS_EC2_M5A_IMG) + ";";
                                    } else
                                    if ( ec2instance.instanceType.startsWith("m5") ) {
                                        prefix = 'shape=image;image=' + dataURIFor(AWS_EC2_M5_IMG) + ";";
                                    } else
                                    if ( ec2instance.instanceType.startsWith("p2") ) {
                                        prefix = 'shape=image;image=' + dataURIFor(AWS_EC2_P2_IMG) + ";";
                                    } else
                                    if ( ec2instance.instanceType.startsWith("p3") ) {
                                        prefix = 'shape=image;image=' + dataURIFor(AWS_EC2_P3_IMG) + ";";
                                    } else
                                    if ( ec2instance.instanceType.startsWith("r4") ) {
                                        prefix = 'shape=image;image=' + dataURIFor(AWS_EC2_R4_IMG) + ";";
                                    } else
                                    if ( ec2instance.instanceType.startsWith("r5a") ) {
                                        prefix = 'shape=image;image=' + dataURIFor(AWS_EC2_R5A_IMG) + ";";
                                    } else
                                    if ( ec2instance.instanceType.startsWith("r5") ) {
                                        prefix = 'shape=image;image=' + dataURIFor(AWS_EC2_R5_IMG) + ";";
                                    }
                                    
                                    var id           = ec2instance.instanceId;
                                    var colorInsType = colorForLabel(ec2instance.instanceType);
                                    var colorState   = colorForLabel(ec2instance.state.name.toUpperCase());
                                    var label        = "<b>" + ec2instance.instanceId + "</b><br>" + ec2instance.privateIpAddress + "<br><span style='color:" + colorInsType + ";'>" + ec2instance.instanceType + "</span>";
                                    var v            = graph.insertVertex(parent, id, label, 0, 0, width, height, prefix+'verticalLabelPosition=bottom;verticalAlign=top;whiteSpace=wrap;');
                                    v.data           = new CustomData('ec2', ec2instance.instanceId, json, v0.cellId);
                                    v.modalInfo      = new ModalInfoHTML('ec2',ec2instance);
                                    var e            = graph.insertEdge(parent,null,ec2instance.state.name.toUpperCase(), v0, v, 'strokeColor=' + rgbToHexStr(colorState) + ';');
                                }
                            }
                        } else {
                            console.log("json is not defined");
                        }
                     }
                     finally
                     {
                         // Updates the display
                         graph.getModel().endUpdate();
                     }
         
                     var enc = new mxCodec();
                     var node = enc.encode(graph.getModel());
                     
                     return mxUtils.getXml(node);
                 };

                 function CustomData(awsResourceType, awsResourceId, sourceData, parentCellId) {
                    this.awsResourceType = awsResourceType;
                    this.awsResourceId   = awsResourceId;
                    this.sourceData      = sourceData;
                    this.parentCellId    = parentCellId;
                 }

                 function ModalInfoHTML(resourceId, data) {
                    if ( resourceId == 'root' ) {
                       this.html = buildHtmlAwsModalInfo(data); 
                    } else
                    if ( resourceId == 'vpc' ) {
                       this.html = buildHtmlVpcModalInfo(data);
                    } else
                    if ( resourceId == 'subnet' ) {
                       this.html = buildHtmlSubnetModalInfo(data);   
                    } else
                    if ( resourceId == 'ec2' ) {
                       this.html = buildHtmlEc2ModalInfo(data);
                    } else {
                       this.html = "<h5>" + JSON.stringify(data) + "</h5>"; 
                    }
                 }


                 var tagBehavior = 'onmouseover="javascript:this.style.backgroundColor=' + "'" + 'rgb(0,175,175)' + "'" + ';this.style.cursor=' + "'" + 'pointer' + "'" + ';"onmouseout="javascript:this.style.backgroundColor=' + "'" + 'rgb(0,220,220)' + "';" + '"';
                 function buildHtmlVpcModalInfo(data) {
                     var html =
                         '<div id="imgBtnClose" style="position:absolute;top:7%;left:93%"><img  onMouseOver="this.style.cursor=' + "'" + 'pointer' + "'" + '" src=""" + self.images.close +""" onClick="javascript:closeModalInfo();"></div>'
                       + '<table class="tableInfo" width="100%" height="100%" border="0" cellpadding="0" cellspacing="0">'
                       + ' <tr><td align="center" height="35px" colspan="2" class="cellHeader">VPC</td></tr>'  
                       + ' <tr><td class="cellInfo">'
                       + '     <input type="image" class="buttonCopyClipboard" onclick="javascript:copyToClipboard(' + "'" + data.vpcId + "'" + ');" src=""" + self.images.copyClipboardPath +""" />'
                       + '     Vpc Id.......:&nbsp;<font style="color:blue;">' + data.vpcId + '</font><br>'
                       + '     <input type="image" class="buttonCopyClipboard" onclick="javascript:copyToClipboard(' + "'" + data.cidrBlock + "'"+ ');" src=""" + self.images.copyClipboardPath +""" />'
                       + '     CIDR Block...:&nbsp;<font style="color:blue;">' + data.cidrBlock + '</font>'
                       + ' <tr>'
                       + ' <tr><td colspan="2" class="cellTags">'
                       + '     Tags:<br><br>'
                       var tagsPerLine = 0;
                       for(var index = 0; index < data.tags.length; index++) {
                           var tag = data.tags[index];
                           html = html + '<div class="tag"' 
                                       + 'onclick="javascript:copyToClipboard(' + "'" + tag.key + "=" + tag.value + "'" + ');" '
                                       + tagBehavior + '>' + tag.key + '=' + tag.value + '</div>';
                           tagsPerLine += 1 
                           if (tagsPerLine >= 2) {
                               tagsPerLine = 0;
                               html = html + '<br><br>'
                           }
                       }
                       + ' </td>'
                       + '</table>'
                    return html;  
                 }

                 function buildHtmlSubnetModalInfo(data) {
                     var labelScope = data.scope == "public" ? "PUBLIC" : "PRIVATE";
                     var html =  
                         '<div id="imgBtnClose" style="position:absolute;top:5%;left:93%"><img  onMouseOver="this.style.cursor=' + "'" + 'pointer' + "'" + '" src=""" + self.images.close +""" onClick="javascript:closeModalInfo();"></div>'
                       + '<table class="tableInfo" width="100%" height="100%" border="0" cellpadding="0" cellspacing="0">'
                       + ' <tr><td align="center" height="35px" colspan="2" class="cellHeader">' + labelScope + ' SUBNET</td></tr>'  
                       + ' <tr><td class="cellInfo">'
                       + '     <input type="image" class="buttonCopyClipboard" onclick="javascript:copyToClipboard(' + "'" + data.subnetId + "'" + ');" src=""" + self.images.copyClipboardPath +""" />'
                       + '     Subnet Id.......:&nbsp;<font style="color:blue;">' + data.subnetId + '</font><br>'
                       + '     <input type="image" class="buttonCopyClipboard" onclick="javascript:copyToClipboard(' + "'" + data.availabilitZone + "'" + ');" src=""" + self.images.copyClipboardPath +""" />'
                       + '     AZ..............:&nbsp;<font style="color:blue;">' + data.availabilitZone + ' (' + data.availabilitZoneId  + ')' + '</font><br>'
                       + '     <input type="image" class="buttonCopyClipboard" onclick="javascript:copyToClipboard(' + "'" + data.cidrBlock + "'"+ ');" src=""" + self.images.copyClipboardPath +""" />'
                       + '     CIDR Block......:&nbsp;<font style="color:blue;">' + data.cidrBlock + '</font><br>'
                       + '     <input type="image" class="buttonCopyClipboard" onclick="javascript:copyToClipboard(' + "'" + data.availableIpAddressCount + "'"+ ');" src=""" + self.images.copyClipboardPath +""" />'
                       + '     Available IpV4..:&nbsp;<font style="color:blue;">' + data.availableIpAddressCount + '</font><br>'
                       + '     <input type="image" class="buttonCopyClipboard" onclick="javascript:copyToClipboard(' + "'" + data.vpcId + "'"+ ');" src=""" + self.images.copyClipboardPath +""" />'
                       + '     VPC Id..........:&nbsp;<font style="color:blue;">' + data.vpcId + '</font><br>'
                       + '     <input type="image" class="buttonCopyClipboard" onclick="javascript:copyToClipboard(' + "'" + data.defaultForAz + "'"+ ');" src=""" + self.images.copyClipboardPath +""" />'
                       + '     Default for AZ..:&nbsp;<font style="color:blue;">' + data.defaultForAz + '</font><br>'
                       + '     <input type="image" class="buttonCopyClipboard" onclick="javascript:copyToClipboard(' + "'" + data.subnetArn + "'"+ ');" src=""" + self.images.copyClipboardPath +""" />'
                       + '     Subnet ARN......:&nbsp;<font style="color:blue;font-size:9px;">' + data.subnetArn + '</font>'
                       + ' <tr>'
                       + ' <tr><td colspan="2" class="cellTags">'
                       + '     Tags:<br><br>'
                       var tagsPerLine = 0;
                       for(var index = 0; index < data.tags.length; index++) {
                           var tag = data.tags[index];
                           html = html + '<div class="tag"' 
                                       + 'onclick="javascript:copyToClipboard(' + "'" + tag.key + "=" + tag.value + "'" + ');" '
                                       + tagBehavior + '>' + tag.key + '=' + tag.value + '</div>';
                           tagsPerLine += 1 
                           if (tagsPerLine >= 2) {
                               tagsPerLine = 0;
                               html = html + '<br><br>'
                           }
                       }
                       + ' </td>'
                       + '</table>'
                    return html;  
                 }

                 function thousandSeparator(x) {
                    return x.toString().replace(/\B(?=(\d{3})+(?!\d))/g, ".");
                 }

                 function callBackShowHtmlSecurityGroup(response, event) {
                     //TODO: Build the HTML Popup for Security Group
                     console.log(event);
                     console.log(response);
                 }

                 function buildHtmlEc2ModalInfo(data) {
                     var colorState;
                     if ( data.state.name.toUpperCase() == "RUNNING" ) {
                         colorState = "green";
                     } else
                     if ( data.state.name.toUpperCase() == "STOPPED" ) {
                         colorState = "red";
                     } else {
                         colorState = "black";
                     }
                     var dtToday      = new Date();
                     var dt           = data.launchTime.split("-")[1] + "-" + data.launchTime.split("-")[0] + "-" + data.launchTime.split("-")[2];
                     var dtLaunched   = new Date(dt);
                     var timeLaucnhed = milliseconds = Math.abs(dtLaunched - dtToday);
                     var hours        = "&nbsp;(<font color=gray;>" + thousandSeparator(parseInt((milliseconds / 36e5))) + " hours)</font>";
                     var publicIpAddressEntry = "";
                     if ( data.publicIpAddress && data.publicIpAddress != "" ){
                           publicIpAddressEntry = '     <input type="image" class="buttonCopyClipboard" onclick="javascript:copyToClipboard(' + "'" + data.publicIpAddress + "'" + ');" src=""" + self.images.copyClipboardPath +""" />'
                                                + '     Public IP...........:&nbsp;<font style="color:blue;">' + data.publicIpAddress + '</font><br>';
                     }
                     var blockDevices = "";
                     if ( data.blockDeviceMappings ) {
                          var totalDevices = data.blockDeviceMappings.length
                          for( var i = 0; i < data.blockDeviceMappings.length; i++ ) {
                              var blockDevice = data.blockDeviceMappings[0];
                              var labelTotal  = '<font style="color:gray;">[' + (i+1) + ' of ' + totalDevices + ']</font>';
                              blockDevices =  blockDevices
                              + '     <input type="image" class="buttonCopyClipboard" onclick="javascript:copyToClipboard(' + "'" + blockDevice.ebs.volumeId + "'" + ');" src=""" + self.images.copyClipboardPath +""" />'
                              + '     EBS...............' + labelTotal + '..:&nbsp;<font style="color:blue;">' + blockDevice.deviceName + ' (' + blockDevice.ebs.volumeId +')&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;</font><br>'
                              + '&nbsp;&nbsp;&nbsp;&nbsp;|Status....................:&nbsp;<font style="color:blue;">' + blockDevice.ebs.status + '</font><br>'
                              + '&nbsp;&nbsp;&nbsp;&nbsp;|Attach Time...............:&nbsp;<font style="color:blue;">' + blockDevice.ebs.attachTime + '</font><br>'
                              + '&nbsp;&nbsp;&nbsp;&nbsp;|Delete on Terminatation...:&nbsp;<font style="color:blue;">' + blockDevice.ebs.deleteOnTermination + '</font><br>'  
                          }
                     }
                     var securityGroups = "";
                     if ( data.securityGroups ) {
                          var totalSecurityGroups = data.securityGroups.length
                          for( var i = 0; i < data.securityGroups.length; i++ ) {
                              var securityGroup = data.securityGroups[0];
                              var labelTotal  = '<font style="color:gray;">[' + (i+1) + ' of ' + totalSecurityGroups + ']</font>';
                              securityGroups =  securityGroups
                              + '     <input type="image" class="buttonCopyClipboard" onclick="javascript:copyToClipboard(' + "'" + securityGroup.groupId + "'" + ');" src=""" + self.images.copyClipboardPath +""" />'
                              + '     Security Group....' + labelTotal + '..:&nbsp;<font style="color:darkorange;text-decoration-line:underline;">'
                              + '<span onMouseOver="this.style.cursor=' + "'" + 'pointer' + "'" + '" onClick="javascript:callListSecurityGroups(' + "'" + securityGroup.groupId + "'" + ',callBackShowHtmlSecurityGroup, event);">' + securityGroup.groupId + '</span></font><br>'
                              + '&nbsp;&nbsp;&nbsp;&nbsp;|Group Name................:&nbsp;<font style="color:blue;">' + securityGroup.groupName + '</font><br>'
                          }
                     }
                     var html =  
                         '<div id="imgBtnClose" style="position:absolute;top:4%;left:93%"><img  onMouseOver="this.style.cursor=' + "'" + 'pointer' + "'" + '" src=""" + self.images.close +""" onClick="javascript:closeModalInfo();"></div>'
                       + '<table class="tableInfo" width="100%" height="100%" border="0" cellpadding="0" cellspacing="0">'
                       + ' <tr><td align="center" height="35px" colspan="2" class="cellHeader">EC2 ' + data.instanceType.toUpperCase() + '</td></tr>'  
                       + ' <tr><td class="cellInfo">'
                       + '     <input type="image" class="buttonCopyClipboard" onclick="javascript:copyToClipboard(' + "'" + data.instanceId + "'" + ');" src=""" + self.images.copyClipboardPath +""" />'
                       + '     Instance Id.................:&nbsp;<font style="color:blue;">' + data.instanceId + '</font><br>'
                       + '     <input type="image" class="buttonCopyClipboard" onclick="javascript:copyToClipboard(' + "'" + data.privateIpAddress + "'" + ');" src=""" + self.images.copyClipboardPath +""" />'
                       + '     Private IP..................:&nbsp;<font style="color:blue;">' + data.privateIpAddress + '</font><br>'
                       + '     <input type="image" class="buttonCopyClipboard" onclick="javascript:copyToClipboard(' + "'" + data.launchTime + "'" + ');" src=""" + self.images.copyClipboardPath +""" />'
                       + '     Launch Time.................:&nbsp;<font style="color:blue;">' + data.launchTime + '</font>' + hours + '<br>'
                       + publicIpAddressEntry
                       + '     <input type="image" class="buttonCopyClipboard" onclick="javascript:copyToClipboard(' + "'" + data.state.name.toUpperCase() + "'" + ');" src=""" + self.images.copyClipboardPath +""" />'
                       + '     State.......................:&nbsp;<font style="color:' + colorState +';">' + data.state.name.toUpperCase() + '</font><br>'

                       + '     <input type="image" class="buttonCopyClipboard" onclick="javascript:copyToClipboard(' + "'" + data.cpu.coreCount + "'" + ');" src=""" + self.images.copyClipboardPath +""" />'
                       + '     CPU Core Count..............:&nbsp;<font style="color:blue;">' + data.cpu.coreCount + '</font><br>'
                       + '     <input type="image" class="buttonCopyClipboard" onclick="javascript:copyToClipboard(' + "'" + data.cpu.coreCount + "'" + ');" src=""" + self.images.copyClipboardPath +""" />'
                       + '     CPU Threads Per Core........:&nbsp;<font style="color:blue;">' + data.cpu.threadsPerCore + '</font><br>'
                       + securityGroups
                       + blockDevices
                       + '     <input type="image" class="buttonCopyClipboard" onclick="javascript:copyToClipboard(' + "'" + data.subnetId + "'" + ');" src=""" + self.images.copyClipboardPath +""" />'
                       + '     Subnet Id...................:&nbsp;<font style="color:blue;">' + data.subnetId + '</font><br>'
                       + '     <input type="image" class="buttonCopyClipboard" onclick="javascript:copyToClipboard(' + "'" + data.vpcId + "'"+ ');" src=""" + self.images.copyClipboardPath +""" />'
                       + '     VPC Id......................:&nbsp;<font style="color:blue;">' + data.vpcId + '</font><br>'
                       + ' <tr>'
                       + ' <tr><td colspan="2" class="cellTags">'
                       + '     Tags:<br><br>'
                       var tagsPerLine = 0;
                       for(var index = 0; index < data.tags.length; index++) {
                           var tag = data.tags[index];
                           html = html + '<div class="tag"' 
                                       + 'onclick="javascript:copyToClipboard(' + "'" + tag.key + "=" + tag.value + "'" + ');" '
                                       + tagBehavior + '>' + tag.key + '=' + tag.value + '</div>';
                           tagsPerLine += 1 
                           if (tagsPerLine >= 2) {
                               tagsPerLine = 0;
                               html = html + '<br><br>'
                           }
                       }
                       + ' </td>'
                       + '</table>'
                    return html;  
                 }

                 function buildHtmlAwsModalInfo(data) {
                     var html =  
                         '<table class="tableInfo" width="100%" height="100%" border="0" cellpadding="0" cellspacing="0">'
                       + ' <tr><td align="center" height="35px" colspan="2" class="cellHeader">AWS</td></tr>'  
                       + ' <tr><td class="cellInfo">'
                       + '    <font style="color:blue;">&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;AWS&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;</font><br>'
                       + ' <tr>'
                       + '</table>'
                    return html;  
                 }

                 function copyToClipboard(str) {
                    const el = document.createElement('textarea');
                    el.value = str;
                    document.body.appendChild(el);
                    el.select();
                    document.execCommand('copy');
                    document.body.removeChild(el);
                 }

                 """ + self.htmlSnippets.jsFormatXml() + """
                 function pad2(n) { return n < 10 ? '0' + n : n }

                 function timestamp() {
                    var date = new Date();
                    return  date.getFullYear().toString() + "/" + pad2(date.getMonth() + 1) + "/" +  pad2( date.getDate()) 
                    +  " " + pad2( date.getHours() ) + ":" + pad2( date.getMinutes() ) + ":" + pad2( date.getSeconds() ) + ":" + pad2( date.getMilliseconds() ) + " ";
                 }
             </script>
          </head>   
          """ + self.htmlSnippets.htmlCommonBody() + """
        </html>
        """
        self.openHTML(html)
        self.startHttpServer()
        print(Emoticons.ok() + " \033[93mNavigator exited!\033[0m")    



import socket
import sys

HOST, PORT = "localhost", 9999
data = " ".join(sys.argv[1:])

# Create a socket (SOCK_STREAM means a TCP socket)
with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
    # Connect to server and send data
    sock.connect((HOST, PORT))
    sock.sendall(bytes(data + "\n", "utf-8"))

    # Receive data from the server and shut down
    received = str(sock.recv(1024), "utf-8")

print("Sent:     {}".format(data))
print("Received: {}".format(received))